
import math
import numpy

from ..offrot import fit_offset_and_rotation


def test_off_rot():
    ang = -0.5
    off = numpy.array([0.23, -0.1])
    ca = math.cos(ang)
    sa = math.sin(ang)
    rot = numpy.array([[ca, -sa], [sa, ca]])

    coords0 = numpy.array(
        [[0, 0], [0, 0.5], [0, 0.1], [0, 0.2], [0.5, 0],
         [0.1, 0], [0.3, 0], [0.2, 0]]
    )
    coords1 = numpy.dot(rot, coords0.T).T + off

    off01, rot01 = fit_offset_and_rotation(coords0, coords1)

    assert numpy.allclose(off, off01)
    assert numpy.allclose(rot, rot01)


def test_off_rot_fb():

    ang = -0.5
    ca = math.cos(ang)
    sa = math.sin(ang)
    rot = numpy.array([[ca, -sa], [sa, ca]])
    off = numpy.array([0.23, -0.1])

    coords0 = numpy.array(
        [[0, 0], [0, 0.5], [1.3, 0.1], [-3.0, 0.2], [0.5, 0],
         [0.1, 0], [0.3, 0], [0.2, 0]]
    )
    c0m = coords0.mean(axis=0)
    cb0 = coords0 - c0m
    coords1 = numpy.dot(rot, cb0.T).T + off
    c1m = coords1.mean(axis=0)

    off_b = -numpy.dot(rot, c0m.T).T + off
    off_a = -numpy.dot(rot.T, c1m.T).T + c0m

    off1, rot1 = fit_offset_and_rotation(coords0, coords1)
    assert numpy.allclose(off1, off_b)
    assert numpy.allclose(rot, rot1)

    off2, rot2 = fit_offset_and_rotation(coords1, coords0)
    assert numpy.allclose(off2, off_a)
    assert numpy.allclose(numpy.transpose(rot), rot2)


def test_off_rot_182():
    p2 = [(1444.96997070312, 468.980010986328),
          (1352.80004882812, 1517.91003417969)]
    p1 = numpy.array(p2)

    q2 = numpy.array(
        [[1445.99857531, 458.24017547],
         [1353.24937729, 1507.67677833]]
    )
    q1 = numpy.array(q2)

    off1, rot1 = fit_offset_and_rotation(p1, q1)
    assert numpy.allclose(off1, [1.24147529, -11.19375707])

    off2, rot2 = fit_offset_and_rotation(q1, p1)
    assert numpy.allclose(off2, [-1.23581507, 11.19438338])

    assert numpy.allclose(rot1, rot2.T)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Compute the solid angle in arcsec^2 for each pixel in a 2D image."""

import argparse
from argparse import RawTextHelpFormatter
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS
import numpy as np
from tqdm import tqdm
from scipy.ndimage import median_filter
from spherical_geometry.polygon import SphericalPolygon
import sys

from .add_script_info_to_fits_history import add_script_info_to_fits_history
from .ctext import ctext


def pixel_solid_angle_arcsec2(wcs, naxis1, naxis2, method=3, kernel_size=None, verbose=False):
    """Compute the solid angle (arcsec**2) of every pixel.

    This function computes the solid angle for each pixel in a 2D image.
    When using WCS with a very small projected pixel size (e.g., 0.01 arcsec),
    Method 2 is recommended but slow. Method 1 is slow and does not
    work well for very small pixel sizes, but we keep the code here for completeness.
    Method 3 is fast and works well for most cases, but it may not be accurate
    for very small pixel sizes.

    Parameters
    ----------
    wcs : `astropy.wcs.WCS`
        WCS object defining the celestial coordinates of the image.
    naxis1 : int
        Number of pixels along the first axis (NAXIS1).
    naxis2 : int
        Number of pixels along the second axis (NAXIS2).
    method : int, optional
        Method to compute the solid angle.
        1: Use spherical polygons (slow).
        2: Use spherical polygons with a different approach (slow).
        3: Use spherical coordinates and distances (fast, default).
    kernel_size : int, optional
        Size of the kernel for smoothing the result using a median filter.
        If None, no smoothing is applied.
    verbose : bool, optional
        If True, display intermediate information.

    Returns
    -------
    result : `numpy.ndarray`
        2D array with the solid angle (arcsec**2) for each pixel in
        the input image.
    """

    if verbose:
        print(f"Computing solid angle for {naxis1} x {naxis2} pixels using method {method}.")

    # X, Y coordinates (2D image array, following the FITS criterium)
    # corresponding to the four corners of all the image pixels
    borders_x = np.arange(naxis1 + 1) + 0.5
    borders_y = np.arange(naxis2 + 1) + 0.5
    meshgrid = np.meshgrid(borders_x, borders_y)
    ix_array = meshgrid[0].flatten()
    iy_array = meshgrid[1].flatten()

    # spherical coordinates of the four corners of all the image pixels
    result_spherical = SkyCoord.from_pixel(
        xp=ix_array,
        yp=iy_array,
        wcs=wcs,
        origin=1,
        mode='all'
    )

    if method == 1:
        # Use spherical polygons to compute the solid angle
        result_spherical = result_spherical.reshape(naxis2 + 1, naxis1 + 1)
        result = np.zeros((naxis2, naxis1))
        for i in tqdm(range(naxis2), desc="NAXIS2", disable=not verbose):
            for j in range(naxis1):
                polygon = SphericalPolygon.from_radec(
                    lon=[result_spherical[i, j].ra.rad,
                         result_spherical[i, j + 1].ra.rad,
                         result_spherical[i + 1, j + 1].ra.rad,
                         result_spherical[i + 1, j].ra.rad],
                    lat=[result_spherical[i, j].dec.rad,
                         result_spherical[i, j + 1].dec.rad,
                         result_spherical[i + 1, j + 1].dec.rad,
                         result_spherical[i + 1, j].dec.rad],
                    degrees=False,
                )
                result[i, j] = polygon.area() * (180 / np.pi) ** 2 * 3600 ** 2
    elif method == 2:
        # Use spherical polygons with a different approach to compute the solid angle
        result_spherical = result_spherical.reshape(naxis2 + 1, naxis1 + 1)
        result = np.zeros((naxis2, naxis1))
        for i in tqdm(range(naxis2), desc="NAXIS2", disable=not verbose):
            for j in range(naxis1):
                dist1a = result_spherical[i, j].separation(result_spherical[i, j + 1]).arcsec
                dist1b = result_spherical[i + 1, j].separation(result_spherical[i + 1, j + 1]).arcsec
                dist2a = result_spherical[i, j].separation(result_spherical[i + 1, j]).arcsec
                dist2b = result_spherical[i, j + 1].separation(result_spherical[i + 1, j + 1]).arcsec
                # average distances corresponding to opposite sides of each pixel
                dist1 = (dist1a + dist1b) / 2
                dist2 = (dist2a + dist2b) / 2
                # solid angle (arcsec**2) of the pixel
                result[i, j] = dist1 * dist2
    elif method == 3:
        # cartesian coordinates of the four corners of all the image pixels
        x = result_spherical.cartesian.x.value.reshape(naxis2 + 1, naxis1 + 1)
        y = result_spherical.cartesian.y.value.reshape(naxis2 + 1, naxis1 + 1)
        z = result_spherical.cartesian.z.value.reshape(naxis2 + 1, naxis1 + 1)
        # dot product of consecutive points along NAXIS1
        dot_product_naxis1 = x[:, :-1] * x[:, 1:] + \
            y[:, :-1] * y[:, 1:] + z[:, :-1] * z[:, 1:]
        # distance (arcsec) between consecutive points along NAXIS1
        result_naxis1 = np.arccos(dot_product_naxis1) * 180 / np.pi * 3600
        # average distances corresponding to the upper and lower sides of each pixel
        pixel_size_naxis1 = (result_naxis1[:-1, :] + result_naxis1[1:, :]) / 2
        # dot product of consecutive points along NAXIS2
        dot_product_naxis2 = x[:-1, :] * x[1:, :] + \
            y[:-1, :] * y[1:, :] + z[:-1, :] * z[1:, :]
        # distance (arcsec) between consecutive points along NAXIS2
        result_naxis2 = np.arccos(dot_product_naxis2) * 180 / np.pi * 3600
        # average distances corresponding to the left and right sides of each pixel
        pixel_size_naxis2 = (result_naxis2[:, :-1] + result_naxis2[:, 1:]) / 2
        # pixel size (arcsec**2)
        result = pixel_size_naxis1 * pixel_size_naxis2
    else:
        raise ValueError(f"Invalid method: {method}. Choose 1, 2, or 3.")

    # smooth result
    if kernel_size is not None:
        if verbose:
            print(f"Applying median filter with kernel size {kernel_size}.")
        result = median_filter(result, size=kernel_size, mode='nearest')

    return result


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Compute the solid angle in arcsec^2 for each pixel in a 2D image.",
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('input_file', type=str,
                        help='FITS file containing the image data.')
    parser.add_argument('output_file', type=str,
                        help='Output FITS file to save the solid angle data.')
    parser.add_argument("--extname", type=str,
                        help="Extension name of the input HDU (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--method", type=int, default=3,
                        help="Method to compute the solid angle:\n"
                             "1: Use spherical polygons (slow)\n"
                             "   (not recommended for very small pixel sizes)\n"
                             "2: Use spherical polygons (different approach, slow)\n"
                             "   (recommended for small pixel sizes)\n"
                             "3: Use spherical coordinates and distances (fast)\n"
                             "   (default, not recommended for very small pixel sizes)")
    parser.add_argument("--kernel_size", type=int, default=None,
                        help="Kernel size for smoothing the result using a median\n"
                             "filter. If not specified, no smoothing is applied.")
    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    input_file = args.input_file
    output_file = args.output_file
    extname = args.extname
    kernel_size = args.kernel_size
    method = args.method
    verbose = args.verbose

    with fits.open(input_file) as hdul:
        if extname not in hdul:
            raise ValueError(f"Extension '{extname}' not found in {input_file}.")
        hdu_image = hdul[extname]
        naxis = hdu_image.header['NAXIS']
        if naxis == 2:
            wcs = WCS(hdu_image.header)
        elif naxis == 3:
            wcs = WCS(hdu_image.header).celestial
        else:
            raise ValueError(f"Unsupported NAXIS value: {naxis}. Expected 2 or 3.")
        naxis1 = hdu_image.header['NAXIS1']
        naxis2 = hdu_image.header['NAXIS2']
        if args.verbose:
            print(f"{naxis1=}")
            print(f"{naxis2=}")
            print(f"Celestial coordinates WCS:\n{wcs}")

    result = pixel_solid_angle_arcsec2(
        wcs=wcs,
        naxis1=naxis1,
        naxis2=naxis2,
        method=method,
        kernel_size=kernel_size,
        verbose=verbose
    )

    # Create a new FITS HDU with the solid angle data
    solid_angle_hdu = fits.PrimaryHDU(data=result.astype(np.float32))
    header = solid_angle_hdu.header
    add_script_info_to_fits_history(header, args)
    if args.verbose:
        print(f"Saving solid angle data to {output_file}.")
    solid_angle_hdu.writeto(output_file, overwrite=True)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Remove extension from FITS file"""
import argparse
from astropy.io import fits
import sys

from .ctext import ctext


def remove_extension(filename, extension, verbose=True):
    """Remove extension from FITS file

    Parameters
    ----------
    filename : str, file-like or `pathlib.Path`
        FITS filename to be updated.
    extension : str or int
        Extension to remove from FITS file.
        The PRIMARY extension cannot be removed.
    verbose : bool
        If True, display additional information.
    """
    hdul = fits.open(filename, mode='update')
    num_extensions = len(hdul)
    if verbose:
        print(hdul.info())
    if isinstance(extension, int):
        # remove extension by index
        if extension == 0:
            raise ValueError("PRIMARY extension cannot be removed")
        elif extension < len(hdul):
            if verbose:
                print(f"Removing {extension} {hdul[extension].name} from {filename}")
            del hdul[extension]  # programming tip: using hdul.pop(extension) does not work!
        else:
            raise ValueError(f"{extension} is greater than number of available extensions: {num_extensions}")
    elif isinstance(extension, str):
        # remove extension by name
        extname = extension.upper()
        extension_exists = extname in [hdu.name for hdu in hdul]
        if extension_exists:
            for i, hdu in enumerate(hdul):
                if hdu.name == extname:
                    if verbose:
                        print(f"Removing {i} {hdul[extension].name} from {filename}")
                    del hdul[i]   # programming tip: using hdul.pop(i) does not work!
                    break
        else:
            raise ValueError(f"Extension '{extname}' not found")
    else:
        raise ValueError(f"{extension} of type {type(extension)} is not a valid extension")

    if verbose:
        print("Flushing the changes...")

    hdul.flush(output_verify='fix+warn', verbose=verbose)

    if verbose:
        print(hdul.info())

    hdul.close(verbose=verbose)


def main(args=None):
    parser = argparse.ArgumentParser(description="Remmove extension from FITS file")
    parser.add_argument("input", help="Input FITS file", type=str)
    parser.add_argument("--extnum", help="extension number (0=PRIMARY)", type=int, default=None)
    parser.add_argument("--extname", help="extension name", type=str, default=None)
    parser.add_argument("--noverbose", help="do not display additional information", action="store_true")
    parser.add_argument("--echo", help="Display full command line", action="store_true")

    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if not args.noverbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    if args.extnum is None:
        if args.extname is None:
            with fits.open(args.input) as hdul:
                print(hdul.info())
            raise ValueError("You must specify at least one of --extnum and --extname")
        else:
            extension = args.extname
    else:
        if args.extname is None:
            extension = args.extnum
        else:
            raise ValueError(f"{args.extnum} and {args.extname} cannot be used together")

    verbose = not args.noverbose
    remove_extension(args.input, extension, verbose)


if __name__ == "__main__":
    main()




import pytest
import numpy
from ..robustfit import fit_theil_sen


def test__fit_theil_sen(benchmark):
    xfit = numpy.array([57.44273096, 110.78482061, 168.53716769, 195.4526101, 223.12392626, 248.35958438,
                        385.01932674, 408.41698474, 446.01819194, 456.89211403, 486.24413101, 614.73795651,
                        678.44219251, 686.39558972, 772.76291666, 804.33827992, 871.81922467, 907.09872511,
                        1008.61034135, 1019.59354694])
    yfit = numpy.array([4054.10895536, 4105.49508458, 4160.83990688, 4186.98408089, 4213.74617144, 4238.40545978,
                        4370.60513265, 4393.64455769, 4430.47829873, 4440.4094712, 4469.95762253, 4595.21595229,
                        4657.88901648, 4665.91557179, 4751.17748556, 4782.60685178, 4848.55625307, 4883.95460694,
                        4984.3974946, 4995.23118177])
    result_intercept = 3995.4488744
    result_slope = 0.9783033171

    intercept, slope = benchmark(fit_theil_sen, xfit, yfit)

    assert intercept != result_intercept
    assert slope != result_slope


def test_theil_sen_except():

    with pytest.raises(ValueError):
        x = [1, 2]
        y = [3, 4]
        fit_theil_sen(x, y)

    with pytest.raises(ValueError):
        x = [[1, 2, 3, 4, 5, 6, 7, 8]]
        y = [3, 4]
        fit_theil_sen(x, y)

    with pytest.raises(ValueError):
        x = [1, 2, 3, 4, 5, 6, 7.7, 8]
        y = [3, 4]
        fit_theil_sen(x, y)

    with pytest.raises(ValueError):
        x = [1, 2, 3, 4, 5, 6, 7.7, 8]
        y = [[1, 2, 3, 4, 5, 6, 7.7],
             [1, 2, 3, 4, 5, 6, 7.7]]
        fit_theil_sen(x, y)

    with pytest.raises(ValueError):
        x = [[1, 2, 3, 4, 5, 6, 7.7, 8]]
        y = [[[1, 2, 3, 4, 5, 6, 7.7, 8]]]
        fit_theil_sen(x, y)


@pytest.mark.parametrize("n,cols", [(20, 40), (21, 40)])
def test_theil_sen_values2d(n, cols):

    x = numpy.arange(n)
    slopes = numpy.linspace(-500, 500, cols)
    intercepts = numpy.linspace(-20, 20, cols)
    y = slopes * x[:, None] + intercepts

    cin, cslope = fit_theil_sen(x, y)

    assert numpy.allclose(slopes, cslope)
    assert numpy.allclose(intercepts, cin)


@pytest.mark.parametrize("n", [20, 21])
def test_theil_sen_values1d(n):

    x = numpy.arange(n)
    slope = 2.0
    intercept = 17
    y = slope * x + intercept

    cin, cslope = fit_theil_sen(x, y)

    assert numpy.allclose(slope, cslope)
    assert numpy.allclose(intercept, cin)


def test_theil_sen_out_shape():

    n = 10
    x = numpy.arange(n)
    slope = 2.0
    intercept = 17
    y = slope * x + intercept

    res = fit_theil_sen(x, y)
    assert res.shape == (2,)


def test_theil_sen_out_shape2():

    n = 10
    m = 12
    x = numpy.arange(n)
    slope = 2.0
    intercept = 17
    xx = numpy.tile(x, (m, 1)).transpose()
    y = slope * xx + intercept

    res = fit_theil_sen(x, y)
    assert res.shape == (2, m)



#
# Copyright 2013-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Recenter routines"""

import numpy
from scipy.spatial import distance

from numina.array.utils import image_box


# returns y,x
def _centering_centroid_loop(data, center, box):
    # extract raster image
    sl = image_box(center, data.shape, box)
    raster = data[sl]

    # Background estimation for recentering
    background = raster.min()

    braster = raster - background
    threshold = braster.mean()
    mask = braster >= threshold
    if not numpy.any(mask):
        return center, background

    rr = numpy.where(mask, braster, 0)
    norm = rr.sum()
    # All points in thresholded raster are 0.0
    if norm <= 0.0:
        return center, background

    fi, ci = numpy.indices(braster.shape)

    fm = (rr * fi).sum() / norm
    cm = (rr * ci).sum() / norm

    return (fm + sl[0].start, cm + sl[1].start), background


def _centering_centroid_loop_xy(data, center_xy, box):
    center_yx = center_xy[::-1]

    ncenter_yx, back = _centering_centroid_loop(data, center_yx, box)
    ncenter_xy = ncenter_yx[::-1]
    return ncenter_xy, back


def centering_centroid(data, xi, yi, box, nloop=10, toldist=1e-3,
                       maxdist=10.0):
    """
    Computes centroid around point

    Parameters
    ----------
    data
    xi
    yi
    box
    nloop
    toldist
    maxdist

    Returns
    -------
        x, y, background, status, message

        status is:
          * 0: not recentering
          * 1: recentering successful
          * 2: maximum distance reached
          * 3: not converged
    """

    # Store original center
    cxy = (xi, yi)
    origin = (xi, yi)
    # initial background
    back = 0.0

    if nloop == 0:
        return xi, yi, 0.0, 0, 'not recentering'

    for i in range(nloop):
        nxy, back = _centering_centroid_loop_xy(data, cxy, box)
        # _logger.debug('new center is %s', ncenter)
        # if we are to far away from the initial point, break
        dst = distance.euclidean(origin, nxy)
        if dst > maxdist:
            msg = f'maximum distance ({maxdist:5.2f}) from origin reached'
            return cxy[0], cxy[1], back, 2, msg

        # check convergence
        dst = distance.euclidean(nxy, cxy)
        if dst < toldist:
            return nxy[0], nxy[1], back, 1, f'converged in iteration {i}'
        else:
            cxy = nxy

    return nxy[0], nxy[1], back, 3, f'not converged in {nloop} iterations'



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import numpy as np
from numpy.polynomial import Polynomial
from scipy.interpolate import interp1d
from scipy.ndimage.filters import gaussian_filter

from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximplotxy import ximplotxy
from numina.modeling.gaussbox import gauss_box_model


def filtmask(sp, fmin, fmax,
             sp_label='spectrum',
             debugplot=0):
    """Filter spectrum in Fourier space.

    Parameters
    ----------
    sp : numpy array
        Spectrum to be filtered and masked.
    fmin : float
        Minimum frequency to be employed.
    fmax : float
        Maximum frequency to be employed.
    sp_label : str
        Additional string to be attached to labels in legends.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    sp_filt : numpy array
        Filtered spectrum

    """

    # Fourier filtering
    xf = np.fft.fftfreq(sp.size)
    yf = np.fft.fft(sp)
    if abs(debugplot) in (21, 22):
        iok = np.where(xf >= 0)
        ximplotxy(xf[iok], yf[iok].real,
                  plottype='semilogy',
                  xlabel='frequency', ylabel='power',
                  title=sp_label+' (before masking)', debugplot=debugplot)

    cut = (np.abs(xf) > fmax)
    yf[cut] = 0.0
    cut = (np.abs(xf) < fmin)
    yf[cut] = 0.0
    if abs(debugplot) in (21, 22):
        iok = np.where(xf >= 0)
        ximplotxy(xf[iok], yf[iok].real,
                  plottype='semilogy',
                  xlabel='frequency', ylabel='power',
                  title=sp_label+' (after masking)', debugplot=debugplot)

    sp_filt = np.fft.ifft(yf).real
    if abs(debugplot) in (21, 22):
        from numina.array.display.matplotlib_qt import plt
        xdum = np.arange(1, sp_filt.size + 1)
        ax = ximplotxy(xdum, sp, title="filtering " + sp_label,
                       label='original', show=False)
        ax.plot(xdum, sp_filt, label='filtered')
        ax.legend()
        plt.show()

    return sp_filt


def cosinebell(n, frac_cosbell):
    """Return a cosine bell spanning n pixels, masking a fraction of pixels

    Parameters
    ----------
    n : int
        Number of pixels.
    frac_cosbell : float
        Length fraction over which the data will be masked.

    """

    mask = np.ones(n)
    nmasked = int(frac_cosbell * n)
    for i in range(nmasked):
        yval = 0.5 * (1 - np.cos(np.pi * float(i) / float(nmasked)))
        mask[i] = yval
        mask[n - i - 1] = yval

    return mask


def convolve_comb_lines(lines_wave, lines_flux, sigma,
                        crpix1, crval1, cdelt1, naxis1):
    """Convolve a set of lines of known wavelengths and flux.

    Parameters
    ----------
    lines_wave : array like
        Input array with wavelengths
    lines_flux : array like
        Input array with fluxes
    sigma : float
        Sigma of the broadening gaussian to be applied.
    crpix1 : float
        CRPIX1 of the desired wavelength calibration.
    crval1 : float
        CRVAL1 of the desired wavelength calibration.
    cdelt1 : float
        CDELT1 of the desired wavelength calibration.
    naxis1 : integer
        NAXIS1 of the output spectrum.

    Returns
    -------
    xwave : array like
        Array with wavelengths for the output spectrum.
    spectrum : array like
        Array with the expected fluxes at each pixel.

    """

    # generate wavelengths for output spectrum
    xwave = crval1 + (np.arange(naxis1) + 1 - crpix1) * cdelt1

    # initialize output spectrum
    spectrum = np.zeros(naxis1)

    # convolve each line
    for wave, flux in zip(lines_wave, lines_flux):
        sp_tmp = gauss_box_model(x=xwave, amplitude=flux, mean=wave,
                                 stddev=sigma)
        spectrum += sp_tmp

    return xwave, spectrum


def periodic_corr1d(sp_reference, sp_offset,
                    remove_mean=False,
                    frac_cosbell=None,
                    zero_padding=None,
                    fminmax=None,
                    naround_zero=None,
                    norm_spectra=False,
                    nfit_peak=7,
                    sp_label='spectrum',
                    plottitle=None,
                    pdf=None,
                    debugplot=0):
    """Periodic correlation between two spectra, implemented using FFT.

    Parameters
    ----------
    sp_reference : numpy array
        Reference spectrum.
    sp_offset : numpy array
        Spectrum which offset is going to be measured relative to the
        reference spectrum.
    remove_mean : bool
        If True, remove mean from input spectra.
    frac_cosbell : float or None
        Fraction of spectrum where the cosine bell falls to zero.
    zero_padding : int or None
        Number of extended pixels set to zero.
    fminmax : tuple of floats or None
        Minimum and maximum frequencies to be used. If None, no
        frequency filtering is employed.
    naround_zero : int
        Half width of the window (around zero offset) to look for
        the correlation peak. If None, the whole correlation
        spectrum is employed. Otherwise, the peak will be sought
        in the interval [-naround_zero, +naround_zero].
    norm_spectra : bool
        If True, the filtered spectra are normalized before computing
        the correlation function. This can be important when comparing
        the peak value of this function using different spectra.
    nfit_peak : int
        Total number of points (must be odd!) around the peak of the
        crosscorrelation function to be employed to estimate the peak
        location (using a fit to a second order polynomial).
    sp_label : str
        Additional string to be attached to labels in legends.
    plottitle : str
        Optional plot title.
    pdf : PdfFile object or None
        If not None, output is sent to PDF file.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    offset : float
        Offset between the two input spectra.
    fpeak : float
        Maximum of the cross-correlation function.

    """

    # protections
    if sp_reference.ndim != 1 or sp_offset.ndim != 1:
        raise ValueError("Invalid array dimensions")
    if sp_reference.shape != sp_offset.shape:
        raise ValueError("x and y shapes are different")
    if nfit_peak % 2 == 0:
        nfit_peak += 1

    if plottitle is None:
        plottitle = ' '

    # remove mean
    if remove_mean:
        sp_reference_m = sp_reference - np.mean(sp_reference)
        sp_offset_m = sp_offset - np.mean(sp_offset)
    else:
        sp_reference_m = sp_reference.copy()
        sp_offset_m = sp_offset.copy()

    # cosine bell
    if frac_cosbell is not None:
        if frac_cosbell < 0.0 or frac_cosbell > 0.5:
            raise ValueError(f'Invalid frac_cosbell: {frac_cosbell}')
        sp_reference_mb = sp_reference_m * cosinebell(
            sp_reference_m.size, frac_cosbell
        )
        sp_offset_mb = sp_offset_m * cosinebell(
            sp_offset_m.size, frac_cosbell
        )
    else:
        sp_reference_mb = sp_reference_m
        sp_offset_mb = sp_offset_m

    # zero padding
    if zero_padding is not None:
        if zero_padding < 0:
            raise ValueError(f'Invalid zero_padding: {zero_padding}')
        sp_reference_mbz = np.concatenate(
            (sp_reference_mb, np.zeros(zero_padding))
        )
        sp_offset_mbz = np.concatenate(
            (sp_offset_mb, np.zeros(zero_padding))
        )
    else:
        sp_reference_mbz = sp_reference_mb
        sp_offset_mbz = sp_offset_mb

    if fminmax is not None:
        fmin, fmax = fminmax
        sp_reference_filtmask = filtmask(
            sp_reference_mbz,
            fmin=fmin,
            fmax=fmax,
            sp_label='reference ' + sp_label,
            debugplot=debugplot)
        sp_offset_filtmask = filtmask(
            sp_offset_mbz,
            fmin=fmin,
            fmax=fmax,
            sp_label='offset ' + sp_label,
            debugplot=debugplot)
    else:
        sp_reference_filtmask = sp_reference_mbz
        sp_offset_filtmask = sp_offset_mbz

    naxis1 = len(sp_reference_filtmask)
    xcorr = np.arange(naxis1, dtype=int)
    naxis1_half = int(naxis1 / 2)
    naxis1_half_remainder = naxis1 % 2
    for i in range(naxis1_half):
        xcorr[i + naxis1_half + naxis1_half_remainder] -= naxis1
    isort = xcorr.argsort()
    xcorr = xcorr[isort]

    if abs(debugplot) in (21, 22):
        from numina.array.display.matplotlib_qt import plt
        xdum = np.arange(len(sp_reference)) + 1
        xdumf = np.arange(len(sp_reference_filtmask)) + 1
        # reference spectrum
        ax = ximplotxy(xdum, sp_reference, show=False,
                       title='reference ' + sp_label,
                       label='original ' + sp_label)
        if remove_mean or frac_cosbell or zero_padding or fminmax:
            ax.plot(xdumf, sp_reference_filtmask,
                    label='processed ' + sp_label)
        ax.legend()
        plt.show()
        # offset spectrum
        ax = ximplotxy(xdum, sp_offset, show=False,
                       title='offset ' + sp_label,
                       label='original ' + sp_label)
        if remove_mean or frac_cosbell or zero_padding or fminmax:
            ax.plot(xdumf, sp_offset_filtmask,
                    label='processed ' + sp_label)
        ax.legend()
        plt.show()

    if (abs(debugplot) in (21, 22)) or (pdf is not None):
        xdum = np.arange(naxis1) + 1
        ax = ximplotxy(xdum, sp_reference_filtmask, show=False,
                       title=plottitle,
                       label='reference ' + sp_label)
        ax.plot(xdum, sp_offset_filtmask, label='offset ' + sp_label)
        ax.legend()
        if pdf is not None:
            pdf.savefig()
        else:
            pause_debugplot(debugplot=debugplot, pltshow=True)

    # normalize spectra if required
    if norm_spectra:
        sp_reference_norm = np.copy(sp_reference_filtmask)
        sp_offset_norm = np.copy(sp_offset_filtmask)
        sp_dum = np.concatenate((sp_reference_norm, sp_offset_norm))
        spmin = min(sp_dum)
        spmax = max(sp_dum)
        idum = np.where(sp_reference_norm > 0)
        sp_reference_norm[idum] /= spmax
        idum = np.where(sp_reference_norm < 0)
        sp_reference_norm[idum] /= -spmin
        idum = np.where(sp_offset_norm > 0)
        sp_offset_norm[idum] /= spmax
        idum = np.where(sp_offset_norm < 0)
        sp_offset_norm[idum] /= -spmin
        if (abs(debugplot) in (21, 22)) or (pdf is not None):
            xdum = np.arange(naxis1) + 1
            ax = ximplotxy(xdum, sp_reference_norm, show=False,
                           title=plottitle + ' [normalized]',
                           label='reference ' + sp_label)
            ax.plot(xdum, sp_offset_norm, label='offset ' + sp_label)
            ax.legend()
            if pdf is not None:
                pdf.savefig()
            else:
                pause_debugplot(debugplot=debugplot, pltshow=True)
    else:
        sp_reference_norm = sp_reference_filtmask
        sp_offset_norm = sp_offset_filtmask

    corr = np.fft.ifft(np.fft.fft(sp_offset_norm) *
                       np.fft.fft(sp_reference_norm).conj()).real
    corr = corr[isort]

    # determine correlation peak
    if naround_zero is None:
        iminpeak = 0
        imaxpeak = naxis1 - 1
    else:
        izero = np.where(xcorr == 0)[0][0]
        iminpeak = max(izero - naround_zero, 0)
        imaxpeak = min(izero + naround_zero, naxis1 - 1)
    ixpeak = corr[iminpeak:(imaxpeak + 1)].argmax() + iminpeak

    # fit correlation peak with 2nd order polynomial
    nmed = nfit_peak // 2
    imin = ixpeak - nmed
    imax = ixpeak + nmed
    lpeak_ok = True
    if imin < 0 or imax > len(corr):
        x_fit = np.zeros(nfit_peak)
        y_fit = np.zeros(nfit_peak)
        x_refined_peak = 0
        y_refined_peak = 0
        lpeak_ok = False
        poly_peak = Polynomial([0.0])
    else:
        x_fit = xcorr[imin:(imax+1)].astype(float)
        y_fit = corr[imin:(imax+1)]
        poly_peak = Polynomial.fit(x_fit, y_fit, 2)
        poly_peak = Polynomial.cast(poly_peak)
        coef = poly_peak.coef
        if coef[2] != 0:
            x_refined_peak = -coef[1] / (2.0 * coef[2])
        else:
            x_refined_peak = 0.0
        y_refined_peak = poly_peak(x_refined_peak)

    offset = x_refined_peak
    fpeak = y_refined_peak

    if (abs(debugplot) % 10 != 0) or (pdf is not None):
        ax = ximplotxy(xcorr, corr,
                       xlabel='offset (pixels)',
                       ylabel='cross-correlation function',
                       title=plottitle,
                       xlim=(-naxis1/2, naxis1/2), show=False)
        ax.axvline(offset, color='grey', linestyle='dashed')
        coffset = f"(offset:{offset:6.2f} pixels)"
        ax.text(0.01, 0.99, coffset,
                horizontalalignment='left',
                verticalalignment='top',
                transform=ax.transAxes)
        if naround_zero is not None:
            cwindow = f"(peak region: [{-naround_zero},{naround_zero}] pixels)"
            ax.text(0.01, 0.93, cwindow,
                    horizontalalignment='left',
                    verticalalignment='top',
                    transform=ax.transAxes)
        # inset plot
        inset_ax = inset_axes(
            ax,
            width="40%",
            height="40%",
            loc=1
        )
        inset_ax.plot(xcorr, corr)
        if naround_zero is not None:
            inset_ax.set_xlim([-naround_zero, naround_zero])
        else:
            inset_ax.set_xlim([-50, 50])
        if lpeak_ok:
            xplot = np.linspace(xcorr[imin], xcorr[imax], num=50)
            yplot = poly_peak(xplot)
            inset_ax.plot(xplot, yplot, '-')
            inset_ax.plot(x_fit, y_fit, 'o')
            inset_ax.plot([x_refined_peak], [y_refined_peak], 'o')
        inset_ax.axvline(offset, color='grey', linestyle='dashed')
        if pdf is not None:
            pdf.savefig()
        else:
            pause_debugplot(debugplot=debugplot,
                            tight_layout=False, pltshow=True)

    return offset, fpeak


def compute_broadening(wv_obj, sp_obj, wv_ref, sp_ref,
                       sigmalist,
                       remove_mean=False,
                       frac_cosbell=None,
                       zero_padding=None,
                       fminmax=None,
                       naround_zero=None,
                       nfit_peak=None,
                       ax1=None, ax2=None,
                       debugplot=0):
    """Compute broadening to match 'sp_obj' with 'sp_ref'.

    The reference spectrum 'sp_ref' should have a better spectral
    resolution than the object spectrum 'sp_obj'.

    It is assumed that the wavelength arrays are provided in
    ascending order.

    The wavelength sampling 'wv_ref' and 'wv_ref' can be different.

    The comparison is performed in the common wavelength
    interval. Within this interval, both spectra are linearly
    resampled to the result of merging 'wv_obj' and 'wv_ref',
    using for this purpose a constant wavelength increment
    (selected from the minimum wavelength step present in any
    of the two input wavelength arrays).

    Parameters
    ----------
    wv_obj : numpy array
        Wavelength sampling of the object spectrum.
    sp_obj : numpy array
        Flux of the object spectrum.
    wv_ref : numpy array
        Wavelength sampling of the reference spectrum.
    sp_ref : numpy array
        Flux of the reference spectrum.
    sigmalist : numpy array or list
        Sigma broadening (in pixels).
    remove_mean : bool
        If True, remove mean from input spectra.
    frac_cosbell : float or None
        Fraction of spectrum where the cosine bell falls to zero.
    zero_padding : int or None
        Number of extended pixels set to zero.
    fminmax : tuple of floats or None
        Minimum and maximum frequencies to be used. If None, no
        frequency filtering is employed.
    naround_zero : int or None
        Half width of the window (around zero offset) to look for
        the correlation peak.
    nfit_peak : int
        Total number of points (must be odd!) around the peak of the
        crosscorrelation function to be employed to estimate the peak
        location (using a fit to a second order polynomial).
    ax1 : matplotlib Axes object
        If not none, this plot represents the offset and fpeak
        variation as a function of sigma.
    ax2 : matplotlib Axes object
        If not none, this plot represents the two input spectra and
        the broadened reference spectrum.
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.

    Returns
    -------
    offset_broad : float
        Offset (in pixels) between the two input spectra measured
        at the sigma value where fpeak is maximum.
    sigma_broad : float
        Sigma (in pixels) for which fpeak is maximum.
    sp_ref_broad : numpy array
        Reference spectrum broadened with sigma_broad (at the original
        wavelength sampling).

    """

    # determine wether there is an intersection between the two
    # wavelength ranges
    minmax_obj = [min(wv_obj), max(wv_obj)]
    minmax_ref = [min(wv_ref), max(wv_ref)]
    minimum_max = min(minmax_obj[1], minmax_ref[1])
    maximum_min = max(minmax_obj[0], minmax_ref[0])
    overlap = minimum_max - maximum_min
    if overlap <= 0:
        raise ValueError("Wavelength ranges of the two spectra don't overlap!")

    # merge wavelength ranges (removing duplicates)
    wv = np.unique(np.sort(np.concatenate((wv_obj, wv_ref))))
    # truncate to common interval
    wv = wv[(wv >= maximum_min) * (wv <= minimum_max)]
    # check whether the sampling is constant
    delta_wv = list(set(np.diff(wv)))
    if np.allclose(delta_wv, delta_wv[0], rtol=1e-05, atol=1e-08):
        # nothing to do
        pass
    else:
        # linear resampling using the smallest step in any of the input arrays
        min_delta_wv_obj = min(list(set(np.diff(wv_obj))))
        min_delta_wv_ref = min(list(set(np.diff(wv_ref))))
        min_delta_wv = min(min_delta_wv_obj, min_delta_wv_ref)
        wv_min = wv[0]
        wv_max = wv[-1]
        nsamples = int((wv_max - wv_min) / min_delta_wv + 0.5) + 1
        wv = np.linspace(wv_min, wv_max, num=nsamples)
        if abs(debugplot) >= 10:
            print('WARNING (compute_broadening): Interpolation required:')
            print(f'  wv_min..: {wv_min}')
            print(f'  wv_max..: {wv_max}')
            print(f'  step....: {min_delta_wv}')
            print(f'  nsamples: {nsamples}')

    # linear interpolation of input spectrum using the merged
    # wavelength sampling
    funinterp_obj = interp1d(wv_obj, sp_obj, kind='linear')
    funinterp_ref = interp1d(wv_ref, sp_ref, kind='linear')
    flux_obj = funinterp_obj(wv)
    flux_ref = funinterp_ref(wv)

    # normalize each spectrum dividing by its median
    flux_obj /= np.median(flux_obj)
    flux_ref /= np.median(flux_ref)

    # plot initial resampled spectra
    if abs(debugplot) in (21, 22):
        ax = ximplotxy(wv, flux_ref,
                       xlabel='wavelength (Angstrom)',
                       ylabel='flux (arbitrary units)',
                       label='flux_ref', show=False)
        ax.plot(wv, flux_obj, label='flux_obj')
        ax.legend()
        pause_debugplot(debugplot=debugplot, pltshow=True)

    nsigmas = len(list(sigmalist))
    offset = np.zeros(nsigmas)
    fpeak = np.zeros(nsigmas)
    for i, sigma in enumerate(sigmalist):
        # broaden reference spectrum
        flux_ref_broad = gaussian_filter(flux_ref, sigma)
        # plot the two spectra
        if abs(debugplot) in (21, 22):
            ax = ximplotxy(wv, flux_ref_broad,
                           xlabel='wavelength (Angstrom)',
                           ylabel='flux (arbitrary units)',
                           label='flux_ref', show=False)
            ax.plot(wv, flux_obj, label='flux_obj')
            ax.set_title('sigma: ' + str(sigma) + ' pixels')
            ax.legend()
            pause_debugplot(debugplot=debugplot, pltshow=True)
        # periodic correlation between the two spectra
        offset[i], fpeak[i] = periodic_corr1d(
            flux_ref_broad, flux_obj,
            remove_mean=remove_mean,
            frac_cosbell=frac_cosbell,
            zero_padding=zero_padding,
            fminmax=fminmax,
            naround_zero=naround_zero,
            nfit_peak=nfit_peak,
            norm_spectra=True,
            debugplot=debugplot
        )

    if ax1 is not None:
        ax1.plot(sigmalist, offset,
                 color='C0', marker='o', linestyle='', label='offset')
        ax1.set_xlabel('sigma (pixels)')
        ax1.set_ylabel('offset (pixels)', color='C0')
        ax1_ = ax1.twinx()
        ax1_.plot(sigmalist, fpeak,
                  color='C1', marker='o', linestyle='', label='fpeak')
        ax1_.set_ylabel('fpeak', color='C1')

    offset_broad = offset[np.argmax(fpeak)]
    sigma_broad = sigmalist[np.argmax(fpeak)]
    sp_ref_broad = gaussian_filter(sp_ref, sigma_broad)

    if ax2 is not None:
        ax2.plot(wv_obj, sp_obj, label='sp_obj')
        ax2.plot(wv_ref, sp_ref, color='#aaaaaa', label='sp_ref')
        ax2.plot(wv_ref, sp_ref_broad, label='sp_ref_broad')
        ax2.set_xlabel('wavelength (Angstrom)')
        ax2.set_ylabel('flux (arbitrary units)')
        ax2.legend()

    return offset_broad, sigma_broad, sp_ref_broad



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Least squares 2D image fitting to a polynomial."""

import numpy


def _powers(order):
    for rank in range(order + 1):
        for bi in range(rank + 1):
            ai = rank - bi
            yield ai, bi


def _compute_value(power, wg):
    """Return the weight corresponding to single power."""
    if power not in wg:
        p1, p2 = power
        # y power
        if p1 == 0:
            yy = wg[(0, -1)]
            wg[power] = numpy.power(yy, p2 / 2).sum() / len(yy)
        # x power
        else:
            xx = wg[(-1, 0)]
            wg[power] = numpy.power(xx, p1 / 2).sum() / len(xx)
    return wg[power]


def _compute_weight(powers, wg):
    """Return the weight corresponding to given powers."""
    # split
    pow1 = (powers[0], 0)
    pow2 = (0, powers[1])

    cal1 = _compute_value(pow1, wg)
    cal2 = _compute_value(pow2, wg)

    return cal1 * cal2


def imsurfit(data, order, output_fit=False):
    """Fit a bidimensional polynomial to an image.

    :param data: a bidimensional array
    :param integer order: order of the polynomial
    :param bool output_fit: return the fitted image
    :returns: a tuple with an array with the coefficients of the polynomial terms

    >>> import numpy
    >>> xx, yy = numpy.mgrid[-1:1:100j,-1:1:100j]
    >>> z = 456.0 + 0.3 * xx - 0.9* yy
    >>> imsurfit(z, order=1) #doctest: +NORMALIZE_WHITESPACE
    (array([  4.56000000e+02,   3.00000000e-01,  -9.00000000e-01]),)

    """

    # we create a grid with the same number of points
    # between -1 and 1
    c0 = complex(0, data.shape[0])
    c1 = complex(0, data.shape[1])
    xx, yy = numpy.ogrid[-1:1:c0, -1:1:c1]

    ncoeff = (order + 1) * (order + 2) // 2

    powerlist = list(_powers(order))

    # Array with ncoff x,y moments of the data
    bb = numpy.zeros(ncoeff)
    # Moments
    for idx, powers in enumerate(powerlist):
        p1, p2 = powers
        bb[idx] = (data * xx ** p1 * yy ** p2).sum() / data.size

    # Now computing aa matrix
    # it contains \sum x^a y^b
    # most of the terms are zero
    # only those with a and b even remain
    # aa is symmetric

    x = xx[:, 0]
    y = yy[0]

    # weights are stored so we compute them only once
    wg = {(0, 0): 1, (-1, 0): x**2, (0, -1): y**2}
    # wg[(2,0)] = wg[(-1,0)].sum() / len(x)
    # wg[(0,2)] = wg[(0,-1)].sum() / len(y)

    aa = numpy.zeros((ncoeff, ncoeff))
    for j, ci in enumerate(powerlist):
        for i, ri in enumerate(powerlist[j:]):
            p1 = ci[0] + ri[0]
            p2 = ci[1] + ri[1]
            if p1 % 2 == 0 and p2 % 2 == 0:
                # val = (x ** p1).sum() / len(x) * (y ** p2).sum() / len(y)
                val = _compute_weight((p1, p2), wg)
                aa[j, i+j] = val

    # Making symmetric the array
    aa += numpy.triu(aa, k=1).T

    polycoeff = numpy.linalg.solve(aa, bb)

    if output_fit:
        index = 0
        result = 0
        for o in range(order + 1):
            for b in range(o + 1):
                a = o - b
                result += polycoeff[index] * (xx ** a) * (yy ** b)
                index += 1

        return polycoeff, result

    return (polycoeff,)


def vertex_of_quadratic(coeffs):
    C = coeffs[1]
    D = coeffs[2]
    A = coeffs[3]
    E = coeffs[4]
    B = coeffs[5]

    det = 4 * A * B - E**2
    if det <= 0:
        raise ValueError('quadratic has no maximum')

    xm = -(2*B*C - D*E) / det
    ym = -(2*A*D - C*E) / det
    return xm, ym


class FitOne(object):
    def __init__(self, x, y, z):
        """Fit a plane to a region using least squares."""
        x = x.astype('float')
        y = y.astype('float')
        self.xm = x.mean()
        self.ym = y.mean()
        x -= self.xm
        y -= self.ym

        aa = numpy.zeros(shape=(3, 3))
        aa[0, 0] = 1
        aa[1, 1] = (x*x).mean()
        aa[2, 2] = (y*y).mean()
        aa[1, 2] = aa[2, 1] = (x*y).mean()
        bb = [z.mean(), (x*z).mean(), (y*z).mean()]

        self.pf = numpy.linalg.solve(aa, bb)

    def __call__(self, x, y):
        xf = x.astype('float') - self.xm
        yf = y.astype('float') - self.ym
        return self.pf[0] + self.pf[1] * xf + self.pf[2] * yf



#
# Copyright 2019-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Load files stored in disk based on extension"""

import os
import logging

from numina.util.objimport import import_object


_logger = logging.getLogger(__name__)


# FIXME: this is already implemented, elsewhere
# There should be one-- and preferably only one --obvious way to do it.

def is_fits(filename, **kwargs):
    return filename.endswith('.fits')


def read_fits(filename):
    import numina.types.dataframe as df
    import astropy.io.fits as fits
    return df.DataFrame(frame=fits.open(filename))


def read_fits_later(filename):
    import numina.types.dataframe as df
    return df.DataFrame(filename=os.path.abspath(filename))


def is_json(filename, **kwargs):
    return filename.endswith('.json')


def read_json(filename):
    import json

    with open(filename) as fd:
        base = json.load(fd)
    return read_structured(base)


def is_yaml(filename, **kwargs):
    return filename.endswith('.yaml')


def read_yaml(filename):
    import yaml

    with open(filename) as fd:
        base = yaml.load(fd)
    return read_structured(base)


def read_structured(data):
    if 'type_fqn' in data:
        type_fqn = data['type_fqn']
        cls = import_object(type_fqn)
        obj = cls.__new__(cls)
        obj.__setstate__(data)
        return obj
    return data


def unserial(value):
    checkers = [(is_fits, read_fits_later),
                (is_json, read_json), (is_yaml, read_yaml)]
    if isinstance(value, str):
        for check_type, conv in checkers:
            if check_type(value):
                return conv(value)
        else:
            return value
    else:
        return value



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
from astropy.units import Unit
import astropy.units as u
import numpy as np
import os

from numina.tools.ctext import ctext

from .fapply_atmosphere_transmission import fapply_atmosphere_transmission
from .simulate_constant_photlam import simulate_constant_photlam
from .simulate_delta_lines import simulate_delta_lines
from .simulate_spectrum import simulate_spectrum
from .raise_valueerror import raise_ValueError


def generate_spectrum_for_scene_block(scene_fname, scene_block, faux_dict, wave_unit,
                                      wave_min, wave_max, nphotons, wavelength_sampling,
                                      apply_atmosphere_transmission, wave_transmission, curve_transmission,
                                      rng, naxis1_detector,
                                      verbose, plots):
    """Generate photons for the scene block.

    Parameters
    ----------
    scene_fname : str
        YAML scene file name.
    scene_block : dict
        Dictonary storing a scene block.
    faux_dict : Python dictionary
        File names of auxiliary files:
        - skycalc: table with SKYCALC Sky Model Calculator predictions
        - flatpix2pix: pixel-to-pixel flat field
        - model_ifu2detector: 2D polynomial transformation
          x_ifu, y_ify, wavelength -> x_detector, y_detector
    wave_unit : `~astropy.units.core.Unit`
        Wavelength unit to be used in the scene block.
    wave_min : `~astropy.units.Quantity`
        Minimum wavelength to be used in the scene block.
    wave_max : `~astropy.units.Quantity`
        Maximum wavelength to be used in the scene block.
    nphotons : int
        Number of photons to be generated in the scene block.
    wavelength_sampling : str
        Method to sample the wavelength values. Two options are valid:
        - 'random': the wavelengt of each photon is randomly determined
          using the spectrum shape as the density probability function.
        - 'fixed': the wavelength of each photon is exactly determined
          using the spectrum shape as the density probability function.
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    apply_atmosphere_transmission : bool
        If True, apply atmosphere transmission to simulated photons.
    wave_transmission : `~astropy.units.Quantity`
        Wavelength column of the tabulated transmission curve.
    curve_transmission : `~astropy.units.Quantity`
        Transmission values for the wavelengths given in
        'wave_transmission'.
    naxis1_detector : `~astropy.units.Quantity`
        Detector NAXIS1, dispersion direction.
    verbose : bool
        If True, display additional information.
    plots : bool
        If True, plot intermediate results.

    Returns
    -------
    simulated_wave : '~numpy.ndarray'
        Array containing `nphotons` simulated photons with the
        spectrum requested in the scene block.

    """

    spectrum_type = scene_block['spectrum']['type']
    if spectrum_type == 'delta-lines':
        mandatory_keys = ['filename', 'wave_column', 'flux_column']
        for key in mandatory_keys:
            if key not in scene_block['spectrum']:
                raise_ValueError(f"Expected key '{key}' not found!")
        filename = scene_block['spectrum']['filename']
        wave_column = scene_block['spectrum']['wave_column'] - 1
        flux_column = scene_block['spectrum']['flux_column'] - 1
        if filename[0] == '@':
            # retrieve file name from dictionary of auxiliary
            # file names for the considered instrument
            filename = faux_dict[filename[1:]]
        catlines = np.genfromtxt(filename)
        line_wave = catlines[:, wave_column] * Unit(wave_unit)
        if not np.all(np.diff(line_wave.value) > 0):
            raise_ValueError(f'Wavelength array {line_wave=} is not sorted!')
        line_flux = catlines[:, flux_column]
        simulated_wave = simulate_delta_lines(
            line_wave=line_wave,
            line_flux=line_flux,
            nphotons=nphotons,
            wavelength_sampling=wavelength_sampling,
            rng=rng,
            wmin=wave_min,
            wmax=wave_max,
            plots=plots,
            plot_title=filename
        )
    elif spectrum_type == 'skycalc-radiance':
        faux_skycalc = faux_dict['skycalc']
        with fits.open(faux_skycalc) as hdul:
            skycalc_table = hdul[1].data
        if wave_unit != Unit('nm'):
            print(ctext(f'Ignoring wave_unit: {wave_unit} (assuming {u.nm})', faint=True))
        wave = skycalc_table['lam'] * u.nm
        if not np.all(np.diff(wave.value) > 0):
            raise_ValueError(f'Wavelength array {wave=} is not sorted!')
        flux = skycalc_table['flux']
        flux_type = 'photlam'
        simulated_wave = simulate_spectrum(
            wave=wave,
            flux=flux,
            flux_type=flux_type,
            nphotons=nphotons,
            wavelength_sampling=wavelength_sampling,
            rng=rng,
            wmin=wave_min,
            wmax=wave_max,
            convolve_sigma_km_s=0 * u.km / u.s,
            nbins_histo=naxis1_detector.value,
            plots=plots,
            plot_title=os.path.basename(faux_skycalc),
            verbose=verbose
        )
    elif spectrum_type == 'tabulated-spectrum':
        mandatory_keys = ['filename', 'wave_column', 'flux_column', 'flux_type']
        for key in mandatory_keys:
            if key not in scene_block['spectrum']:
                raise_ValueError(f"Expected key '{key}' not found!")
        filename = scene_block['spectrum']['filename']
        wave_column = scene_block['spectrum']['wave_column'] - 1
        flux_column = scene_block['spectrum']['flux_column'] - 1
        flux_type = scene_block['spectrum']['flux_type']
        if 'redshift' in scene_block['spectrum']:
            redshift = float(scene_block['spectrum']['redshift'])
        else:
            if verbose:
                print(ctext('Assuming redshift: 0', faint=True))
            redshift = 0.0
        if 'convolve_sigma_km_s' in scene_block['spectrum']:
            convolve_sigma_km_s = float(scene_block['spectrum']['convolve_sigma_km_s'])
        else:
            if verbose:
                print(ctext('Assuming convolve_sigma_km_s: 0', faint=True))
            convolve_sigma_km_s = 0.0
        convolve_sigma_km_s *= u.km / u.s
        # read data
        table_data = np.genfromtxt(filename)
        wave = table_data[:, wave_column] * (1 + redshift) * Unit(wave_unit)
        if not np.all(np.diff(wave.value) > 0):
            raise_ValueError(f'Wavelength array {wave=} is not sorted!')
        flux = table_data[:, flux_column]
        simulated_wave = simulate_spectrum(
            wave=wave,
            flux=flux,
            flux_type=flux_type,
            nphotons=nphotons,
            wavelength_sampling=wavelength_sampling,
            rng=rng,
            wmin=wave_min,
            wmax=wave_max,
            convolve_sigma_km_s=convolve_sigma_km_s,
            nbins_histo=naxis1_detector.value,
            plots=plots,
            plot_title=os.path.basename(filename),
            verbose=verbose
        )
    elif spectrum_type == 'constant-flux':
        simulated_wave = simulate_constant_photlam(
            wmin=wave_min,
            wmax=wave_max,
            nphotons=nphotons,
            wavelength_sampling=wavelength_sampling,
            rng=rng
        )
    else:
        simulated_wave = None  # avoid PyCharm warning (not aware of raise ValueError)
        raise_ValueError(f'Unexpected {spectrum_type=} in file {scene_fname}')

    # apply atmosphere transmission
    if apply_atmosphere_transmission:
        fapply_atmosphere_transmission(
            simulated_wave=simulated_wave,
            wave_transmission=wave_transmission,
            curve_transmission=curve_transmission,
            rng=rng,
            verbose=verbose,
            plots=plots
        )

    return simulated_wave



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of FRIDA DRP
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from numina.tools.ctext import ctext


def raise_ValueError(msg):
    """Raise exception showing a coloured message."""
    raise ValueError(ctext(msg, fg='red'))



#
# Copyright 2015-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import argparse

import matplotlib
from astropy.io import fits
from astropy.visualization import ZScaleInterval
import numpy as np
import re

from .matplotlib_qt import set_window_geometry
from .pause_debugplot import pause_debugplot
from .fileinfo import list_fileinfo_from_txt
from .fileinfo import check_extnum
from .overplot_ds9reg import overplot_ds9reg
from ..stats import summary


GLOBAL_ASPECT = 'auto'
GLOBAL_GEOMETRY = '1024,768,0,0'  # xwidth,ywidth,xorigin,yorigin
dum_str = ''  # global variable in function keypress
dum_par = ''  # global variable in function keypress


def check_wavelength_scale(crval1, cdelt1, ctype1, cunit1):
    """Check for wavelength calibration in the X axis.

    Parameters
    ----------
    crval1 : float or None
        CRVAL1 parameter corresponding to wavelength calibration in
        the X direction.
    cdelt1 : float or None
        CDELT1 parameter corresponding to wavelength calibration in
        the X direction.
    ctype1 : str or None
        CTYPE1 parameter corresponding to wavelength calibration in
        the X direction.
    cunit1 : str or None
        CUNIT1 parameter corresponding to wavelength calibration in
        the X direction.

    Returns
    -------
    result : bool
        True in the wavelength calibration has been set.
        False otherwise.

    """
    result = False

    if ctype1 is None and cunit1 is None:
        return result

    if ctype1 is not None:
        if 'wavelength' in ctype1.lower():
            result = True
    if cunit1 is not None:
        if 'angstrom' in cunit1.lower():
            result = True

    if result:
        if crval1 is not None and cdelt1 is not None:
            pass
        else:
            result = False

    return result


def ximshow_jupyter(image2d, **args):
    """Auxiliary function to call ximshow from a jupyter notebook.
    """
    return ximshow(image2d, using_jupyter=True, **args)


def ximshow(image2d, title=None, show=True,
            cbar_label=None, cbar_orientation='None',
            z1z2=None, cmap='hot',
            image_bbox=None, first_pixel=(1, 1),
            aspect=GLOBAL_ASPECT,
            crpix1=None, crval1=None, cdelt1=None, ctype1=None, cunit1=None,
            ds9regfile=None,
            geometry=GLOBAL_GEOMETRY, figuredict=None,
            tight_layout=True,
            debugplot=0, using_jupyter=False):
    """Auxiliary function to display a numpy 2d array.

    Parameters
    ----------
    image2d : 2d numpy array, float
        2d image to be displayed.
    title : string
        Plot title.
    cbar_label : string
        Color bar label.
    cbar_orientation : string
        Color bar orientation: valid options are 'horizontal' or
        'vertical' (or 'None' for no color bar).
    show : bool
        If True, the function shows the displayed image. Otherwise
        the function just invoke the plt.imshow() function and
        plt.show() is expected to be executed outside.
    z1z2 : tuple of floats, string or None
        Background and foreground values. If None, zcuts are employed.
    cmap : string
        Color map to be employed.
    image_bbox : tuple (4 integers) or str
        Image rectangle to be displayed, with indices given by
        (nc1,nc2,ns1,ns2), which correspond to the numpy array:
        image2d[(ns1-1):ns2,(nc1-1):nc2].
    first_pixel : tuple (2 integers)
        (x0,y0) coordinates of pixel at origin.
    aspect : str
        Control de aspect ratio of the axes. Valid values are 'equal'
        and 'auto'.
    crpix1 : float or None
        CRPIX1 parameter corresponding to wavelength calibration in
        the X direction.
    crval1 : float or None
        CRVAL1 parameter corresponding to wavelength calibration in
        the X direction.
    cdelt1 : float or None
        CDELT1 parameter corresponding to wavelength calibration in
        the X direction.
    ctype1 : str or None
        CTYPE1 parameter corresponding to wavelength calibration in
        the X direction.
    cunit1 : str or None
        CUNIT1 parameter corresponding to wavelength calibration in
        the X direction.
    ds9regfile : file handler
        Ds9 region file to be overplotted.
    geometry : str or None
        xwidth, ywidth, xorigin, yorigin values employed to set the
        window geometry.
    tight_layout : bool
        If True, and show=True, a tight display layout is set.
    figuredict: dictionary
        Parameters for ptl.figure(). Useful for pdf output.
        For example: --figuredict "{'figsize': (8, 10), 'dpi': 100}"
    debugplot : int
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot
    using_jupyter : bool
        If True, this function is called from a jupyter notebook.

    Returns
    -------
    ax : axes object
        Matplotlib axes instance. This value is returned only when
        'show' is False.

    """

    from numina.array.display.matplotlib_qt import plt

    if not show and using_jupyter:
        plt.ioff()

    # protections
    if not isinstance(image2d, np.ndarray):
        raise ValueError("image2d=" + str(image2d) +
                         " must be a numpy.ndarray")
    elif image2d.ndim != 2:
        raise ValueError("image2d.ndim=" + str(image2d.dim) +
                         " must be 2")

    naxis2_, naxis1_ = image2d.shape

    # check if wavelength calibration is provided
    wavecalib = check_wavelength_scale(
        crval1=crval1, cdelt1=1, ctype1=ctype1, cunit1=cunit1
    )

    # read bounding box limits
    if image_bbox is None:
        nc1 = 1
        nc2 = naxis1_
        ns1 = 1
        ns2 = naxis2_
    elif isinstance(image_bbox, str):
        tmp_bbox = image_bbox.split(",")
        nc1 = int(tmp_bbox[0])
        nc2 = int(tmp_bbox[1])
        ns1 = int(tmp_bbox[2])
        ns2 = int(tmp_bbox[3])
        if nc1 < 1:
            nc1 = 1
        if nc2 > naxis1_:
            nc2 = naxis1_
        if ns1 < 1:
            ns1 = 1
        if ns2 > naxis2_:
            ns2 = naxis2_
    else:
        nc1, nc2, ns1, ns2 = image_bbox
        if 1 <= nc1 <= nc2 <= naxis1_:
            pass
        else:
            raise ValueError("Invalid bounding box limits")
        if 1 <= ns1 <= ns2 <= naxis2_:
            pass
        else:
            raise ValueError("Invalid bounding box limits")

    def get_current_zoom(ax_image, debug=False):
        """Return subimage corresponding to current zoom.

        Parameters
        ----------
        ax_image : axes
            Image axes.
        debug : bool
            If True, the image corners are printed.

        Returns
        -------
        subimage : numpy array (floats)
            Subimage.

        """

        xmin_image, xmax_image = ax_image.get_xlim()
        ymin_image, ymax_image = ax_image.get_ylim()
        ixmin = int(xmin_image + 0.5)
        ixmax = int(xmax_image + 0.5)
        iymin = int(ymin_image + 0.5)
        iymax = int(ymax_image + 0.5)
        ixmin -= first_pixel[0] - 1
        ixmax -= first_pixel[0] - 1
        iymin -= first_pixel[1] - 1
        iymax -= first_pixel[1] - 1
        if ixmin < nc1:
            ixmin = nc1
        if ixmin > nc2:
            ixmin = nc2
        if ixmax < nc1:
            ixmax = nc1
        if ixmax > nc2:
            ixmax = nc2
        if iymin < ns1:
            iymin = ns1
        if iymin > ns2:
            iymin = ns2
        if iymax < ns1:
            iymax = ns1
        if iymax > ns2:
            iymax = ns2
        if debug:
            print("\n>>> xmin, xmax, ymin, ymax [pixels; origin (1,1)]:",
                  ixmin, ixmax, iymin, iymax)
        return image2d[(iymin-1):iymax, (ixmin-1):ixmax]

    def keypress(event):
        """Deal with keyboard events, allowing the update of vmin and vmax.

        Note that a call to raw_input() is not allowed within this
        function since, in that case, the following runtime error
        is raised: can't re-enter readline

        For that reason, the new vmin and vmax values should be
        entered blindly.

        To avoid collisions with navigation keyboard shortcuts,
        check the table available at:
        http://matplotlib.org/users/navigation_toolbar.html

        """

        global dum_str
        global dum_par
        if event.key == "?":
            print("""
Keyword events
==============
Home/Reset......................: h or r or home
Back............................: c or left arrow or backspace
Forward.........................: v or right arrow
Pan/Zoom........................: p
Zoom-to-rect....................: o
Save............................: ctrl + s
Toggle fullscreen...............: ctrl + f
Close plot......................: ctrl + w
Set zscale......................: /
Set bg=min and fg=max values....: ,
Display statistical summary.....: ;
Set foreground by keyboard......: m
Set background by keyboard......: n
Activate/deactivate ds9 regions.: a
Change aspect ratio.............: =
Constrain pan/zoom to x axis....: hold x when panning/zooming with mouse
Constrain pan/zoom to y axis....: hold y when panning/zooming with mouse
Preserve aspect ratio...........: hold CONTROL when panning/zooming with mouse
Toggle grid.....................: g when mouse is over an axes
Toggle x axis scale (log/linear): L or k when mouse is over an axes
Toggle y axis scale (log/linear): l when mouse is over an axes
            """)
        elif event.key == "/":
            subimage2d = get_current_zoom(ax, debug=True)
            new_vmin, new_vmax = ZScaleInterval().get_limits(subimage2d)
            print(">>> setting cuts to vmin=" + str(new_vmin) +
                  " and vmax=" + str(new_vmax))
            im_show.set_clim(vmin=new_vmin)
            im_show.set_clim(vmax=new_vmax)
            dum_str = ''
            dum_par = ''
            plt.show(block=False)
            plt.pause(0.001)
        elif event.key == ",":
            subimage2d = get_current_zoom(ax, debug=True)
            new_vmin = np.nanmin(subimage2d)
            new_vmax = np.nanmax(subimage2d)
            print(">>> setting cuts to vmin=" + str(new_vmin) +
                  " and vmax=" + str(new_vmax))
            im_show.set_clim(vmin=new_vmin)
            im_show.set_clim(vmax=new_vmax)
            dum_str = ''
            dum_par = ''
            plt.show(block=False)
            plt.pause(0.001)
        elif event.key == ";":
            subimage2d = get_current_zoom(ax, debug=True)
            summary(subimage2d.flatten(), debug=True)
        elif event.key == "n":
            print("Type (blindly!) vmin <return>")
            dum_str = ''
            dum_par = "vmin"
        elif event.key == "m":
            print("Type (blindly!) vmax <return>")
            dum_str = ''
            dum_par = "vmax"
        elif event.key == "=":
            if ax.get_aspect() == 'equal':
                ax.set_aspect('auto')
            else:
                ax.set_aspect('equal')
            plt.tight_layout()
            plt.show(block=False)
            plt.pause(0.001)
        elif event.key == "enter":
            if dum_par == "vmin":
                try:
                    new_vmin = float(dum_str)
                except ValueError:
                    print("Invalid vmin=" + dum_str)
                    dum_str = ''
                    print("Type again (blindly!) vmin <return>")
                else:
                    print("Setting vmin=" + dum_str)
                    im_show.set_clim(vmin=new_vmin)
                    dum_str = ''
                    dum_par = ''
                    plt.show(block=False)
                    plt.pause(0.001)
            elif dum_par == "vmax":
                try:
                    new_vmax = float(dum_str)
                except ValueError:
                    print("Invalid vmax=" + dum_str)
                    dum_str = ''
                    print("Type again (blindly!) vmax <return>")
                else:
                    print("Setting vmax=" + dum_str)
                    im_show.set_clim(vmax=new_vmax)
                    dum_str = ''
                    dum_par = ''
                    plt.show(block=False)
                    plt.pause(0.001)
        else:
            if dum_str == '':
                dum_str = event.key
            else:
                dum_str += event.key

    # plot limits
    xmin = float(nc1) - 0.5 + (first_pixel[0] - 1)
    xmax = float(nc2) + 0.5 + (first_pixel[0] - 1)
    ymin = float(ns1) - 0.5 + (first_pixel[1] - 1)
    ymax = float(ns2) + 0.5 + (first_pixel[1] - 1)

    # display image
    if figuredict is None:
        fig = plt.figure()
    else:
        fig = plt.figure(**figuredict)

    ax = fig.add_subplot(111)
    ax.autoscale(False)
    ax.set_xlabel('image pixel in the X direction')
    ax.set_ylabel('image pixel in the Y direction')
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.grid(False)
    if z1z2 in ["zscale", None]:
        z1, z2 = ZScaleInterval().get_limits(
            image2d[(ns1 - 1):ns2, (nc1 - 1):nc2]
        )
    elif z1z2 == "minmax":
        z1 = np.nanmin(image2d[(ns1 - 1):ns2, (nc1 - 1):nc2])
        z2 = np.nanmax(image2d[(ns1 - 1):ns2, (nc1 - 1):nc2])
    else:
        z1, z2 = z1z2
    im_show = plt.imshow(image2d[(ns1 - 1):ns2, (nc1 - 1):nc2],
                         cmap=cmap, aspect=aspect,
                         vmin=z1, vmax=z2,
                         interpolation='nearest', origin='lower',
                         extent=[xmin, xmax, ymin, ymax])
    if cbar_label is None:
        cbar_label = "Number of counts"
    if cbar_orientation in ["horizontal", "vertical"]:
        plt.colorbar(im_show, shrink=1.0, label=cbar_label,
                     orientation=cbar_orientation)
    if title is not None:
        ax.set_title(title)

    if ds9regfile is not None:
        overplot_ds9reg(ds9regfile.name, ax)

    # set the geometry
    if geometry is not None:
        tmp_str = geometry.split(",")
        xwidth_geom = int(tmp_str[0])
        ywidth_geom = int(tmp_str[1])
        xini_geom = int(tmp_str[2])
        yini_geom = int(tmp_str[3])
        backend = matplotlib.get_backend()
        if backend == 'TkAgg':
            plt.get_current_fig_manager().resize(xwidth_geom, ywidth_geom)
            plt.get_current_fig_manager().window.wm_geometry(
                f"+{xini_geom}+{yini_geom}")
        elif backend == 'Qt5Agg':
            geometry_tuple = xwidth_geom, ywidth_geom, xini_geom, yini_geom
            set_window_geometry(geometry_tuple)
        else:
            print(
                f'WARNING: geometry {geometry} ignored with backend {backend}')
            print('(you can try the TkAgg backend instead)')

    # connect keypress event with function responsible for
    # updating vmin and vmax
    fig.canvas.mpl_connect('key_press_event', keypress)

    # wavelength scale
    if wavecalib:
        if crpix1 is None:
            crpix1 = 1.0
        xminwv = crval1 + (xmin - crpix1) * cdelt1
        xmaxwv = crval1 + (xmax - crpix1) * cdelt1
        ax2 = ax.twiny()
        ax2.grid(False)
        ax2.set_xlim(xminwv, xmaxwv)
        ax2.set_xlabel('Wavelength (Angstroms)')
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

    # show plot or return axes
    if show:
        pause_debugplot(debugplot, pltshow=show, tight_layout=tight_layout)
    else:
        if tight_layout:
            plt.tight_layout()
        # return axes
        if using_jupyter:
            plt.ion()
        return ax


def ximshow_file(singlefile,
                 extnum=1,
                 args_cmap=None,
                 args_cbar_label=None, args_cbar_orientation='None',
                 args_z1z2=None, args_bbox=None, args_firstpix=None,
                 args_aspect=GLOBAL_ASPECT,
                 args_keystitle=None, args_ds9reg=None,
                 args_geometry=GLOBAL_GEOMETRY, pdf=None,
                 args_figuredict=None,
                 show=True,
                 debugplot=None,
                 using_jupyter=False):
    """Function to execute ximshow() as called from command line.

    Parameters
    ----------
    singlefile : string
        Name of the FITS file to be displayed.
    extnum : int
        Extension number: 1 for first extension (default).
    args_cmap : string
        Matplotlib color map.
    args_cbar_label : string
        Color bar label.
    args_cbar_orientation : string
        Color bar orientation: valid options are 'horizontal' or
        'vertical' (or 'None' for no color bar).
    args_z1z2 : string or None
        String providing the image cuts tuple: z1, z2, minmax, zscale
        or None
    args_bbox : string or None
        String providing the bounding box tuple: nc1, nc2, ns1, ns2
    args_firstpix : string or None
        String providing the coordinates of lower left pixel.
        args_aspect : str
    args_aspect : str
        Control de aspect ratio of the axes. Valid values are 'equal'
        and 'auto'.
    args_keystitle : string or None
        Tuple of FITS keywords.format: key1,key2,...,keyn.format
    args_ds9reg : file handler
        Ds9 region file to be overplotted.
    args_geometry : string or None
        xwidth, ywidth, xorigin, yorigin to define the window geometry.
        This information is ignored if args_pdffile is not None.
    pdf : PdfFile object or None
        If not None, output is sent to PDF file.
    args_figuredict : string containing a dictionary
        Parameters for ptl.figure(). Useful for pdf output.
        For example: --figuredict "{'figsize': (8, 10), 'dpi': 100}"
    show : bool
        If True, the function shows the displayed image. Otherwise
        the function just invoke the plt.imshow() function and
        plt.show() is expected to be executed outside.
    debugplot : integer or None
        Determines whether intermediate computations and/or plots
        are displayed. The valid codes are defined in
        numina.array.display.pause_debugplot.
    using_jupyter : bool
        If True, this function is called from a jupyter notebook.

    Returns
    -------
    ax : axes object
        Matplotlib axes instance. This value is returned only when
        'show' is False.

    """

    # read z1, z2
    if args_z1z2 in ["zscale", "None", None]:
        z1z2 = None
    elif args_z1z2 == "minmax":
        z1z2 = "minmax"
    else:
        if args_z1z2[0] == '[':
            tmp_str = args_z1z2[1:]
        else:
            tmp_str = args_z1z2
        tmp_str = re.sub(']', '', tmp_str)
        tmp_str = tmp_str.split(",")
        z1z2 = float(tmp_str[0]), float(tmp_str[1])

    # read input FITS file
    hdulist = fits.open(singlefile)
    if extnum is None or extnum < 1 or extnum > len(hdulist):
        raise ValueError(f'Unexpected extension number {extnum}')
    image_header = hdulist[extnum - 1].header
    image2d = hdulist[extnum - 1].data
    hdulist.close()

    naxis1 = image_header['naxis1']
    if 'naxis2' in image_header:
        naxis2 = image_header['naxis2']
    else:
        naxis2 = 1

    # read wavelength calibration
    if 'crpix1' in image_header:
        crpix1 = image_header['crpix1']
    else:
        crpix1 = None
    if 'crval1' in image_header:
        crval1 = image_header['crval1']
    else:
        crval1 = None
    if 'cdelt1' in image_header:
        cdelt1 = image_header['cdelt1']
    else:
        cdelt1 = None
    if 'ctype1' in image_header:
        ctype1 = image_header['ctype1']
    else:
        ctype1 = None
    if 'cunit1' in image_header:
        cunit1 = image_header['cunit1']
    else:
        cunit1 = None

    # title for plot
    title = singlefile
    if args_keystitle is not None:
        keystitle = args_keystitle
        keysformat = ".".join(keystitle.split(".")[1:])
        keysnames = keystitle.split(".")[0]
        tuple_of_keyval = ()
        for key in keysnames.split(","):
            keyval = image_header[key]
            tuple_of_keyval += (keyval,)
        title += "\n" + str(keysformat % tuple_of_keyval)

    if len(image2d.shape) == 1:
        if image2d.shape != (naxis1,):
            raise ValueError("Unexpected error with NAXIS1")
        image2d = np.reshape(image2d, (1, naxis1))
    elif len(image2d.shape) == 2:
        if image2d.shape != (naxis2, naxis1):
            raise ValueError("Unexpected error with NAXIS1, NAXIS2")
    else:
        raise ValueError("Unexpected number of dimensions > 2")

    print('>>> File..:', singlefile)
    print('>>> NAXIS1:', naxis1)
    print('>>> NAXIS2:', naxis2)

    # read bounding box
    if args_bbox is None:
        nc1 = 1
        nc2 = naxis1
        ns1 = 1
        ns2 = naxis2
    else:
        tmp_bbox = args_bbox.split(",")
        nc1 = int(tmp_bbox[0])
        nc2 = int(tmp_bbox[1])
        ns1 = int(tmp_bbox[2])
        ns2 = int(tmp_bbox[3])
        if nc1 < 1:
            nc1 = 1
        if nc2 > naxis1:
            nc2 = naxis1
        if ns1 < 1:
            ns1 = 1
        if ns2 > naxis2:
            ns2 = naxis2

    # read coordinates of lower left pixel
    if args_firstpix is None:
        nc0 = 1
        ns0 = 1
    else:
        tmp_firstpix = args_firstpix.split(",")
        nc0 = int(tmp_firstpix[0])
        ns0 = int(tmp_firstpix[1])

    # display image
    if args_figuredict is None:
        figuredict = None
    else:
        figuredict = eval(args_figuredict)

    ax = ximshow(image2d=image2d, show=False,
                 cmap=args_cmap,
                 cbar_label=args_cbar_label,
                 cbar_orientation=args_cbar_orientation,
                 title=title,
                 z1z2=z1z2,
                 image_bbox=(nc1, nc2, ns1, ns2),
                 first_pixel=(nc0, ns0),
                 aspect=args_aspect,
                 crpix1=crpix1,
                 crval1=crval1,
                 cdelt1=cdelt1,
                 ctype1=ctype1,
                 cunit1=cunit1,
                 ds9regfile=args_ds9reg,
                 geometry=args_geometry,
                 figuredict=figuredict,
                 debugplot=debugplot,
                 using_jupyter=using_jupyter)

    if pdf is not None:
        if show:
            from numina.array.display.matplotlib_qt import plt
            plt.tight_layout()
            pdf.savefig()
        else:
            return ax
    else:
        if show:
            pause_debugplot(debugplot, pltshow=True)
        else:
            # return axes
            return ax


def jimshow(image2d,
            ax=None,
            title=None,
            vmin=None, vmax=None,
            image_bbox=None,
            aspect=GLOBAL_ASPECT,
            xlabel='image pixel in the X direction',
            ylabel='image pixel in the Y direction',
            crpix1=None, crval1=None, cdelt1=None, ctype1=None, cunit1=None,
            grid=False,
            cmap='hot',
            cbar_label='Number of counts',
            cbar_orientation='horizontal'):
    """Auxiliary function to display a numpy 2d array via axes object.

    Parameters
    ----------
    image2d : 2d numpy array, float
        2d image to be displayed.
    ax : axes object
        Matplotlib axes instance. Note that this value is also
        employed as output.
    title : string
        Plot title.
    vmin : float, 'min', or None
        Background value. If None, the minimum zcut is employed.
    vmax : float, 'max', or None
        Foreground value. If None, the maximum zcut is employed.
    image_bbox : tuple (4 integers)
        Image rectangle to be displayed, with indices given by
        (nc1,nc2,ns1,ns2), which correspond to the numpy array:
        image2d[(ns1-1):ns2,(nc1-1):nc2].
    aspect : str
        Control de aspect ratio of the axes. Valid values are 'equal'
        and 'auto'.
    xlabel : string
        X-axis label.
    ylabel : string
        Y-axis label.
    crpix1 : float or None
        CRPIX1 parameter corresponding to wavelength calibration in
        the X direction.
    crval1 : float or None
        CRVAL1 parameter corresponding to wavelength calibration in
        the X direction.
    cdelt1 : float or None
        CDELT1 parameter corresponding to wavelength calibration in
        the X direction.
    ctype1 : str or None
        CTYPE1 parameter corresponding to wavelength calibration in
        the X direction.
    cunit1 : str or None
        CUNIT1 parameter corresponding to wavelength calibration in
        the X direction.
    grid : bool
        If True, overplot grid.
    cmap : string
        Color map to be employed.
    cbar_label : string
        Color bar label.
    cbar_orientation : string
        Color bar orientation: valid options are 'horizontal' or
        'vertical' (or 'None' for no color bar).

    Returns
    -------
    ax : axes object
        Matplotlib axes instance. Note that this value must also
        be provided as input.

    """

    if ax is None:
        raise ValueError('ax=None is not valid in this function')

    naxis2_, naxis1_ = image2d.shape

    if image_bbox is None:
        nc1, nc2, ns1, ns2 = 1, naxis1_, 1, naxis2_
    else:
        nc1, nc2, ns1, ns2 = image_bbox
    if 1 <= nc1 <= nc2 <= naxis1_:
        pass
    else:
        raise ValueError("Invalid bounding box limits")
    if 1 <= ns1 <= ns2 <= naxis2_:
        pass
    else:
        raise ValueError("Invalid bounding box limits")

    # plot limits
    xmin = float(nc1) - 0.5
    xmax = float(nc2) + 0.5
    ymin = float(ns1) - 0.5
    ymax = float(ns2) + 0.5

    image2d_region = image2d[(ns1 - 1):ns2, (nc1 - 1):nc2]

    if vmin is None or vmax is None:
        z1, z2 = ZScaleInterval().get_limits(image2d_region)
    else:
        z1, z2 = None, None

    if vmin is None:
        vmin = z1
    elif vmin == 'min':
        vmin = np.nanmin(image2d_region)
    if vmax is None:
        vmax = z2
    elif vmax == 'max':
        vmax = np.nanmax(image2d_region)

    im_show = ax.imshow(
        image2d_region,
        cmap=cmap, aspect=aspect, vmin=vmin, vmax=vmax,
        interpolation="nearest", origin="lower",
        extent=[xmin, xmax, ymin, ymax]
    )
    if cbar_orientation in ['horizontal', 'vertical']:
        import matplotlib.pyplot as plt
        plt.colorbar(im_show, shrink=1.0,
                     label=cbar_label, orientation=cbar_orientation,
                     ax=ax)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.grid(grid)
    if title is not None:
        ax.set_title(title)

    wavecalib = check_wavelength_scale(
        crval1=crval1, cdelt1=1, ctype1=ctype1, cunit1=cunit1
    )
    if wavecalib:
        if crpix1 is None:
            crpix1 = 1.0
        xminwv = crval1 + (xmin - crpix1) * cdelt1
        xmaxwv = crval1 + (xmax - crpix1) * cdelt1
        ax2 = ax.twiny()
        ax2.grid(False)
        ax2.set_xlim(xminwv, xmaxwv)
        ax2.set_xlabel('Wavelength (Angstroms)')


def jimshowfile(filename,
                extnum=1,
                ax=None,
                title=None,
                vmin=None, vmax=None,
                image_bbox=None,
                aspect=GLOBAL_ASPECT,
                xlabel='image pixel in the X direction',
                ylabel='image pixel in the Y direction',
                crpix1=None, crval1=None, cdelt1=None, ctype1=None, cunit1=None,
                grid=False,
                cmap='hot',
                cbar_label='Number of counts',
                cbar_orientation='horizontal'):
    """Auxiliary function to display a FITS image via axes object.

    Parameters
    ----------
    filename : string
        Input FITS file name.
    extnum : int
        Extension number (1: primary)
    ax : axes object
        Matplotlib axes instance. Note that this value is also
        employed as output.
    title : string
        Plot title.
    vmin : float, 'min', or None
        Background value. If None, the minimum zcut is employed.
    vmax : float, 'max', or None
        Foreground value. If None, the maximum zcut is employed.
    image_bbox : tuple (4 integers)
        Image rectangle to be displayed, with indices given by
        (nc1,nc2,ns1,ns2), which correspond to the numpy array:
        image2d[(ns1-1):ns2,(nc1-1):nc2].
    aspect : str
        Control de aspect ratio of the axes. Valid values are 'equal'
        and 'auto'.
    xlabel : string
        X-axis label.
    ylabel : string
        Y-axis label.
    crpix1 : float or None
        CRPIX1 parameter corresponding to wavelength calibration in
        the X direction.
    crval1 : float or None
        CRVAL1 parameter corresponding to wavelength calibration in
        the X direction.
    cdelt1 : float or None
        CDELT1 parameter corresponding to wavelength calibration in
        the X direction.
    ctype1 : str or None
        CTYPE1 parameter corresponding to wavelength calibration in
        the X direction.
    cunit1 : str or None
        CUNIT1 parameter corresponding to wavelength calibration in
        the X direction.
    grid : bool
        If True, overplot grid.
    cmap : string
        Color map to be employed.
    cbar_label : string
        Color bar label.
    cbar_orientation : string
        Color bar orientation: valid options are 'horizontal' or
        'vertical' (or 'None' for no color bar).

    Returns
    -------
    ax : axes object
        Matplotlib axes instance. Note that this value must also
        be provided as input.

    """

    # read input FITS file
    hdulist = fits.open(filename)
    if extnum is None or extnum < 1 or extnum > len(hdulist):
        raise ValueError(f'Unexpected extension number {extnum}')
    image2d = hdulist[extnum - 1].data
    hdulist.close()

    return jimshow(image2d,
                   ax=ax,
                   title=title,
                   vmin=vmin, vmax=vmax,
                   image_bbox=image_bbox,
                   aspect=aspect,
                   xlabel=xlabel,
                   ylabel=ylabel,
                   crpix1=crpix1, crval1=crval1, cdelt1=cdelt1,
                   ctype1=ctype1, cunit1=cunit1,
                   grid=grid,
                   cmap=cmap,
                   cbar_label=cbar_label,
                   cbar_orientation=cbar_orientation)


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: display FITS images'
    )

    # positional arguments
    parser.add_argument("filename",
                        help="FITS file (wildcards allowed) "
                             "or txt file with list of FITS files",
                        nargs="+")

    # optional arguments
    parser.add_argument('--extnum',
                        help='Extension number in input files (note that ' +
                             'first extension is 1 = default value)',
                        default=1, type=int)
    parser.add_argument("--z1z2",
                        help="tuple [z1,z2], minmax, zscale or None (=zscale)",
                        type=str)
    parser.add_argument("--bbox",
                        help="bounding box tuple: nc1,nc2,ns1,ns2")
    parser.add_argument("--firstpix",
                        help="coordinates of lower left pixel: nc0, ns0")
    parser.add_argument("--aspect",
                        help="aspect ratio (equal or auto)",
                        type=str,
                        choices=['equal', 'auto'], default=GLOBAL_ASPECT)
    parser.add_argument("--cmap",
                        help="color map",
                        type=str, default='hot')
    parser.add_argument("--cbar_label",
                        help="color bar label",
                        type=str, default='Number of counts')
    parser.add_argument("--cbar_orientation",
                        help="color bar orientation",
                        type=str,
                        choices=['horizontal', 'vertical', 'None'],
                        default='horizontal')
    parser.add_argument("--keystitle",
                        help="tuple of FITS keywords.format: " +
                             "key1,key2,...keyn.'format'")
    parser.add_argument("--ds9reg",
                        help="ds9 region file to be overplotted",
                        type=argparse.FileType('rt'))
    parser.add_argument("--geometry",
                        help='string "xwidth,ywidth,xorigin,yorigin"',
                        default=GLOBAL_GEOMETRY)
    parser.add_argument("--pdffile",
                        help="ouput PDF file name",
                        type=argparse.FileType('w'))
    parser.add_argument("--figuredict",
                        help="string with dictionary of parameters for"
                             "plt.figure()",
                        type=str)
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                             " (default=12)",
                        default=12, type=int,
                        choices=[0, 1, 2, 10, 11, 12, 21, 22])
    args = parser.parse_args(args)

    if abs(args.debugplot) in [21, 22]:
        print('>> args.filename: ', args.filename)

    if len(args.filename) == 1:
        list_fits_files = []
        list_extnum = []
        for tmp in list_fileinfo_from_txt(args.filename[0]):
            list_fits_files.append(tmp.filename)
            list_extnum.append(tmp.extnum)
    else:
        list_fits_files = []
        list_extnum = []
        for tmp in args.filename:
            tmpfile, tmpextnum = check_extnum(tmp)
            for tmptmp in list_fileinfo_from_txt(tmpfile):
                list_fits_files.append(tmptmp.filename)
                list_extnum.append(tmpextnum)

    list_extnum = [args.extnum if dum is None else dum for dum in list_extnum]

    if abs(args.debugplot) in [21, 22]:
        print('>> Filenames.: ', list_fits_files)
        print('>> Extensions: ', list_extnum)

    # read pdffile
    if args.pdffile is not None:
        from matplotlib.backends.backend_pdf import PdfPages
        pdf = PdfPages(args.pdffile.name)
    else:
        from numina.array.display.matplotlib_qt import plt  # noqa: F401
        pdf = None

    for myfile, extnum in zip(list_fits_files, list_extnum):
        if extnum is None:
            extnum = args.extnum
        ximshow_file(singlefile=myfile,
                     extnum=extnum,
                     args_z1z2=args.z1z2,
                     args_bbox=args.bbox,
                     args_firstpix=args.firstpix,
                     args_aspect=args.aspect,
                     args_cmap=args.cmap,
                     args_cbar_label=args.cbar_label,
                     args_cbar_orientation=args.cbar_orientation,
                     args_keystitle=args.keystitle,
                     args_ds9reg=args.ds9reg,
                     args_geometry=args.geometry,
                     pdf=pdf,
                     args_figuredict=args.figuredict,
                     debugplot=args.debugplot)

    if pdf is not None:
        pdf.close()


if __name__ == "__main__":

    main()



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Unit test for taggers"""

import pytest

import astropy.io.fits as fits

from numina.types.dataframe import DataFrame
from ..oresult import ObservationResult
from ..taggers import get_tags_from_full_ob


def test_empty_ob():

    ob = ObservationResult()
    tags = get_tags_from_full_ob(ob)

    assert len(tags) == 0


def test_init_ob():

    img1 = fits.PrimaryHDU(data=[1, 2, 3])
    frame1 = DataFrame(frame=fits.HDUList(img1))

    ob = ObservationResult()
    ob.frames = [frame1]
    tags = get_tags_from_full_ob(ob)

    assert len(tags) == 0


def test_header_key1_ob():

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD1'
    frame1 = DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD2'
    frame2 = DataFrame(frame=fits.HDUList(img1))

    ob = ObservationResult()
    ob.frames = [frame1, frame2]
    tags = get_tags_from_full_ob(ob, reqtags=['FILTER'])

    assert tags == {'FILTER': 'FILTER-A'}


def test_header_key1_mis():

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    frame1 = DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['FILTER'] = 'FILTER-B'
    frame2 = DataFrame(frame=fits.HDUList(img2))

    ob = ObservationResult()
    ob.frames = [frame1, frame2]

    with pytest.raises(ValueError):
        get_tags_from_full_ob(ob, reqtags=['FILTER'])


def test_header_key2_ob():

    img1 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img1.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD1'
    frame1 = DataFrame(frame=fits.HDUList(img1))

    img2 = fits.PrimaryHDU(data=[1, 2, 3], header=fits.Header())
    img2.header['FILTER'] = 'FILTER-A'
    img1.header['READM'] = 'MOD1'
    frame2 = DataFrame(frame=fits.HDUList(img1))

    ob = ObservationResult()
    ob.frames = [frame1, frame2]
    tags = get_tags_from_full_ob(ob, reqtags=['FILTER', 'READM'])

    assert tags == {'FILTER': 'FILTER-A', 'READM': 'MOD1'}



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Double check that __setstate__ and __getstate__ methods work properly."""

import importlib


def check_setstate_getstate(self, json_file_name):
    class_name = self.__class__.__name__

    ClassName = getattr(importlib.import_module('emirdrp.products'),
                        class_name)

    # concatenate __getstate__ and __setstate__
    instance = ClassName()
    instance.__setstate__(self.__getstate__())
    instance.writeto(json_file_name + '_bis')

    # load data from input JSON file and save them again
    instance_bis = ClassName._datatype_load(json_file_name)
    instance_bis.writeto(json_file_name + '_bis2')



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np
from numpy.polynomial import polynomial

from ..arccalibration import gen_triplets_master
from ..arccalibration import arccalibration_direct
from ..arccalibration import fit_list_of_wvfeatures

try:
    import matplotlib  # noqa: F401

    HAVE_PLOTS = True
except ImportError:
    HAVE_PLOTS = False


# -----------------------------------------------------------------------------


def simulate_master_table(my_seed, wv_ini_master, wv_end_master, nlines_master,
                          ldebug=False):
    """Generates a simulated master table of wavelengths.

    The location of the lines follows a random uniform distribution
    between `wv_ini_master` and `wv_end_master`. The total number of generated
    lines is `nlines_master`. The seed for random number generation can be
    re-initialized with `my_seed`.

    Parameters
    ----------
    my_seed : int
        Seed to re-initialize random number generation.
    wv_ini_master : float
        Minimum wavelength in master table.
    wv_end_master : float
        Maximum wavelength in master table.
    nlines_master : int
        Total number of lines in master table.
    ldebug : bool
        If True intermediate results are displayed.

    Returns
    -------
    wv_master : numpy array
        Array with wavelengths corresponding to the master table (Angstroms).
    """

    if my_seed is not None:
        np.random.seed(my_seed)

    if wv_end_master < wv_ini_master:
        raise ValueError('wv_ini_master=' + str(wv_ini_master) +
                         ' must be <= wv_end_master=' + str(wv_end_master))

    wv_master = np.random.uniform(low=wv_ini_master,
                                  high=wv_end_master,
                                  size=nlines_master)
    wv_master.sort()  # in-place sort

    if ldebug:
        print('>>> Master table:')
        for val in zip(range(nlines_master), wv_master):
            print(val)

    return wv_master


# -----------------------------------------------------------------------------


def simulate_arc(wv_ini_master, wv_end_master, wv_master,
                 wv_ini_arc, wv_end_arc, naxis1_arc,
                 prob_line_master_in_arc,
                 delta_xpos_min_arc, delta_lambda, error_xpos_arc,
                 poly_degree, fraction_unknown_lines,
                 ldebug=False, lplot=False):
    """Generates simulated input for arc calibration.

    Parameters
    ----------
    wv_ini_master : float
        Minimum wavelength in master table.
    wv_end_master : float
        Maximum wavelength in master table.
    wv_master : numpy array
        Array with wavelengths corresponding to the master table (Angstroms).
    wv_ini_arc : float
        Minimum wavelength in arc spectrum.
    wv_end_arc : float
        Maximum wavelength in arc spectrum.
    naxis1_arc : int
        NAXIS1 of arc spectrum.
    prob_line_master_in_arc : float
        Probability that a master table line appears in the arc spectrum.
    delta_xpos_min_arc : float
        Minimum distance (pixels) between lines in arc spectrum.
    delta_lambda : float
        Maximum deviation (Angstroms) from linearity in arc calibration
        polynomial.
    error_xpos_arc : float
        Standard deviation (pixels) employed to introduce noise in the arc
        spectrum lines. The initial lines are shifted from their original
        location following a Normal distribution with mean iqual to zero and
        sigma equal to this parameter.
    poly_degree : int
        Polynomial degree corresponding to the original wavelength calibration
        function.
    fraction_unknown_lines : float
        Fraction of lines that on average will be unknown (i.e., lines that
        appear in the arc spectrum that are not present in the master table).
    ldebug : bool
        If True intermediate results are displayed.
    lplot : bool
        If True intermediate plots are displayed.

    Returns
    -------
    nlines_arc : int
        Number of arc lines
    xpos_arc : numpy array
        Location of arc lines (pixels).
    crval1_arc : float
        CRVAL1 for arc spectrum (linear approximation).
    cdelt1_arc : float
        CDELT1 for arc spectrum (linear approximation).
    c0_arc, c1_arc, c2_arc : floats
        Coefficients of the second order polynomial.
    ipos_wv_arc : numpy array
        Number of line in master table corresponding to each arc line. Unknown
        lines (i.e. those that are not present in the master table) are
        assigned to -1.
    coeff_original : numpy array
        Polynomial coefficients ordered from low to high, corresponding to the
        fit to the arc lines before the inclusion of unknown lines.

    """

    if (wv_ini_arc < wv_ini_master) or (wv_ini_arc > wv_end_master):
        print('wv_ini_master:', wv_ini_master)
        print('wv_end_master:', wv_end_master)
        print('wv_ini_arc...:', wv_ini_arc)
        raise ValueError('wavelength_ini_arc outside valid range')

    if (wv_end_arc < wv_ini_master) or (wv_end_arc > wv_end_master):
        print('wv_ini_master:', wv_ini_master)
        print('wv_end_master:', wv_end_master)
        print('wv_end_arc...:', wv_end_arc)
        raise ValueError('wavelength_ini_arc outside valid range')

    if wv_end_arc < wv_ini_arc:
        raise ValueError('wv_ini_arc=' + str(wv_ini_arc) +
                         ' must be <= wv_end_arc=' + str(wv_end_arc))

    # ---

    # nlines_master = wv_master.size

    crval1_arc = wv_ini_arc
    cdelt1_arc = (wv_end_arc - wv_ini_arc) / float(naxis1_arc - 1)
    crpix1_arc = 1.0
    if ldebug:
        print('>>> CRVAL1, CDELT1, CRPIX1....:', crval1_arc, cdelt1_arc,
              crpix1_arc)

    # ---
    # The arc lines constitute a subset of the master lines in the considered
    # wavelength range.
    i1_master = np.searchsorted(wv_master, wv_ini_arc)
    i2_master = np.searchsorted(wv_master, wv_end_arc)
    nlines_temp = i2_master - i1_master
    nlines_arc_ini = int(round(nlines_temp * prob_line_master_in_arc))
    ipos_wv_arc_ini = np.random.choice(range(i1_master, i2_master),
                                       size=nlines_arc_ini,
                                       replace=False)
    ipos_wv_arc_ini.sort()  # in-place sort
    wv_arc_ini = wv_master[ipos_wv_arc_ini]
    if ldebug:
        print('>>> Number of master lines in arc region.:', nlines_temp)
        print('>>> Initial number of arc lines..........:', nlines_arc_ini)
        print('>>> Initial selection of master list lines for arc:')
        print(ipos_wv_arc_ini)
    # Remove too close lines.
    ipos_wv_arc = np.copy(ipos_wv_arc_ini[0:1])
    wv_arc = np.copy(wv_arc_ini[0:1])
    i_last = 0
    for i in range(1, nlines_arc_ini):
        delta_xpos = (wv_arc_ini[i] - wv_arc_ini[i_last]) / cdelt1_arc
        if delta_xpos > delta_xpos_min_arc:
            ipos_wv_arc = np.append(ipos_wv_arc, ipos_wv_arc_ini[i])
            wv_arc = np.append(wv_arc, wv_arc_ini[i])
            i_last = i
        else:
            if ldebug:
                print('--> skipping line #', i, '. Too close to line #',
                      i_last)
    nlines_arc = len(wv_arc)
    if ldebug:
        print('>>> Intermediate number of arc lines.....:', nlines_arc)
        print('>>> Intermediate selection of master list lines for arc:')
        print(ipos_wv_arc)

    # Generate pixel location of the arc lines.
    if delta_lambda == 0.0:
        # linear solution
        xpos_arc = (wv_arc - crval1_arc) / cdelt1_arc + 1.0
        c0_arc = wv_ini_arc
        c1_arc = cdelt1_arc
        c2_arc = 0.0
    else:
        # polynomial solution
        c0_arc = wv_ini_arc
        c1_arc = (wv_end_arc - wv_ini_arc - 4 * delta_lambda) / float(
            naxis1_arc - 1)
        c2_arc = 4 * delta_lambda / float(naxis1_arc - 1) ** 2
        xpos_arc = (-c1_arc + np.sqrt(c1_arc ** 2 -
                    4 * c2_arc * (c0_arc - wv_arc)))
        xpos_arc /= 2 * c2_arc
        xpos_arc += 1  # convert from 0,...,(NAXIS1-1) to 1,...,NAXIS1

    # Introduce noise in arc line positions.
    if error_xpos_arc > 0:
        xpos_arc += np.random.normal(loc=0.0,
                                     scale=error_xpos_arc,
                                     size=nlines_arc)
    # Check that the order of the lines has not been modified.
    xpos_arc_copy = np.copy(xpos_arc)
    xpos_arc_copy.sort()  # in-place sort
    if sum(xpos_arc == xpos_arc_copy) != len(xpos_arc):
        raise ValueError(
            'FATAL ERROR: arc line switch after introducing noise')

    if lplot and HAVE_PLOTS:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlim([1, naxis1_arc])
        ax.set_ylim([wv_ini_arc, wv_end_arc])
        ax.plot(xpos_arc, wv_arc, 'ro')
        xp = np.array([1, naxis1_arc])
        yp = np.array([wv_ini_arc, wv_end_arc])
        ax.plot(xp, yp, 'b-')
        xp = np.arange(1, naxis1_arc + 1)
        yp = c0_arc + c1_arc * (xp - 1) + c2_arc * (xp - 1) ** 2
        ax.plot(xp, yp, 'g:')
        ax.set_xlabel('pixel position in arc spectrum')
        ax.set_ylabel('wavelength (Angstrom)')
        plt.show(block=False)

    # Unweighted polynomial fit.
    coeff_original = polynomial.polyfit(xpos_arc, wv_arc, poly_degree)
    poly_original = polynomial.Polynomial(coeff_original)

    if lplot and HAVE_PLOTS:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlim([1, naxis1_arc])
        if delta_lambda == 0.0:
            if error_xpos_arc > 0:
                ax.set_ylim([-4 * error_xpos_arc * cdelt1_arc,
                             4 * error_xpos_arc * cdelt1_arc])
            else:
                ax.set_ylim([-1.1, 1.1])
        else:
            ax.set_ylim([-delta_lambda * 1.5, delta_lambda * 1.5])
        yp = wv_arc - (crval1_arc + (xpos_arc - 1) * cdelt1_arc)
        ax.plot(xpos_arc, yp, 'ro')
        xp = np.array([1, naxis1_arc])
        yp = np.array([0, 0])
        ax.plot(xp, yp, 'b-')
        xp = np.arange(1, naxis1_arc + 1)
        yp = c0_arc + c1_arc * (xp - 1) + c2_arc * (xp - 1) ** 2
        yp -= crval1_arc + cdelt1_arc * (xp - 1)
        ax.plot(xp, yp, 'g:')
        yp = poly_original(xp)
        yp -= crval1_arc + cdelt1_arc * (xp - 1)
        ax.plot(xp, yp, 'm-')
        ax.set_xlabel('pixel position in arc spectrum')
        ax.set_ylabel('residuals (Angstrom)')
        plt.show(block=False)

    # ---
    # Include unknown lines (lines that do not appear in the master table).
    nunknown_lines = int(round(fraction_unknown_lines * float(nlines_arc)))
    if ldebug:
        print('>>> Number of unknown arc lines..........:', nunknown_lines)
    for i in range(nunknown_lines):
        iiter = 0
        while True:
            iiter += 1
            if iiter > 1000:
                raise ValueError('iiter > 1000 while adding unknown lines')
            xpos_dum = np.random.uniform(low=1.0,
                                         high=float(naxis1_arc),
                                         size=1)
            isort = np.searchsorted(xpos_arc, xpos_dum)
            newlineok = False
            if isort == 0:
                dxpos1 = abs(xpos_arc[isort] - xpos_dum)
                if dxpos1 > delta_xpos_min_arc:
                    newlineok = True
            elif isort == nlines_arc:
                dxpos2 = abs(xpos_arc[isort - 1] - xpos_dum)
                if dxpos2 > delta_xpos_min_arc:
                    newlineok = True
            else:
                dxpos1 = abs(xpos_arc[isort] - xpos_dum)
                dxpos2 = abs(xpos_arc[isort - 1] - xpos_dum)
                if (dxpos1 > delta_xpos_min_arc) and \
                        (dxpos2 > delta_xpos_min_arc):
                    newlineok = True
            if newlineok:
                xpos_arc = np.insert(xpos_arc, isort, xpos_dum)
                ipos_wv_arc = np.insert(ipos_wv_arc, isort, -1)
                nlines_arc += 1
                if ldebug:
                    print('--> adding unknown line at pixel:', xpos_dum)
                break
    if ldebug:
        print('>>> Final number of arc lines............:', nlines_arc)
        for val in zip(range(nlines_arc), ipos_wv_arc, xpos_arc):
            print(val)

    if lplot and HAVE_PLOTS:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_ylim([0.0, 3.0])
        ax.vlines(wv_master, 0.0, 1.0)
        ax.vlines(wv_arc, 1.0, 2.0, colors='r', linestyle=':')
        ax.vlines(wv_ini_arc, 0.0, 3.0, colors='m', linewidth=3.0)
        ax.vlines(wv_end_arc, 0.0, 3.0, colors='m', linewidth=3.0)
        ax.set_xlabel('wavelength')
        axbis = ax.twiny()
        axbis.vlines(xpos_arc, 2.0, 3.0, colors='g')
        xmin_xpos_master = (wv_ini_master - crval1_arc) / cdelt1_arc + 1.0
        xmax_xpos_master = (wv_end_master - crval1_arc) / cdelt1_arc + 1.0
        axbis.set_xlim([xmin_xpos_master, xmax_xpos_master])
        axbis.set_xlabel('pixel position in arc spectrum')
        plt.show(block=False)

    return nlines_arc, xpos_arc, crval1_arc, cdelt1_arc, c0_arc, c1_arc, c2_arc, ipos_wv_arc, coeff_original


# -----------------------------------------------------------------------------

def execute_arccalibration(my_seed=432, wv_ini_master=3000, wv_end_master=7000,
                           nlines_master=120,
                           wv_ini_arc=4000, wv_end_arc=5000,
                           naxis1_arc=1024, crpix1=1.0,
                           prob_line_master_in_arc=0.80,
                           delta_xpos_min_arc=4.0,
                           delta_lambda=5.0, error_xpos_arc=0.3,
                           poly_degree=2, fraction_unknown_lines=0.20,
                           wv_ini_search=None, wv_end_search=None,
                           times_sigma_r=3.0, frac_triplets_for_sum=0.50,
                           times_sigma_theil_sen=10.0, poly_degree_wfit=2,
                           times_sigma_polfilt=10.0,
                           times_sigma_cook=10.0,
                           times_sigma_inclusion=5.0,
                           ldebug=False, lplot=False):
    """Execute a particular arc calibration simulation.

    This function simulates a master list, generates a simulated arc, and
    carry out its wavelength calibration.

    Parameters
    ----------
    my_seed : int
        Seed to re-initialize random number generation.
    wv_ini_master : float
        Minimum wavelength in master table.
    wv_end_master : float
        Maximum wavelength in master table.
    nlines_master : int
        Total number of lines in master table.
    ldebug : bool
        If True intermediate results are displayed.

    Returns
    -------
    coeff : numpy array
        Coefficients of the polynomial fit.
    crval1_approx : float
        Approximate CRVAL1 value.
    cdetl1_approx : float
        Approximate CDELT1 value.
    """

    wv_master = simulate_master_table(my_seed, wv_ini_master, wv_end_master,
                                      nlines_master,
                                      ldebug=ldebug)
    ntriplets_master, ratios_master_sorted, triplets_master_sorted_list = \
        gen_triplets_master(wv_master)

    nlines_arc, xpos_arc, crval1_arc, cdelt1_arc, c0_arc, c1_arc, c2_arc, ipos_wv_arc, coeff_original = \
        simulate_arc(
            wv_ini_master, wv_end_master, wv_master,
            wv_ini_arc, wv_end_arc, naxis1_arc,
            prob_line_master_in_arc, delta_xpos_min_arc, delta_lambda, error_xpos_arc,
            poly_degree, fraction_unknown_lines, ldebug=ldebug, lplot=lplot
        )

    if wv_ini_search is None:
        wv_ini_search = wv_ini_master - 0.1 * (wv_end_master - wv_ini_master)
    if wv_end_search is None:
        wv_end_search = wv_end_master + 0.1 * (wv_end_master - wv_ini_master)

    list_of_wvfeatures = arccalibration_direct(
        wv_master=wv_master,
        ntriplets_master=ntriplets_master,
        ratios_master_sorted=ratios_master_sorted,
        triplets_master_sorted_list=triplets_master_sorted_list,
        xpos_arc=xpos_arc,
        naxis1_arc=naxis1_arc,
        crpix1=crpix1,
        wv_ini_search=wv_ini_search,
        wv_end_search=wv_end_search,
        wvmin_useful=None,
        wvmax_useful=None,
        error_xpos_arc=error_xpos_arc,
        times_sigma_r=times_sigma_r,
        frac_triplets_for_sum=frac_triplets_for_sum,
        times_sigma_theil_sen=times_sigma_theil_sen,
        poly_degree_wfit=poly_degree_wfit,
        times_sigma_polfilt=times_sigma_polfilt,
        times_sigma_cook=times_sigma_cook,
        times_sigma_inclusion=times_sigma_inclusion)

    solution_wv = \
        fit_list_of_wvfeatures(
            list_of_wvfeatures=list_of_wvfeatures,
            naxis1_arc=naxis1_arc,
            crpix1=crpix1,
            poly_degree_wfit=poly_degree_wfit
        )

    return solution_wv


# -----------------------------------------------------------------------------


def test__execute_notebook_example(ldebug=False, lplot=False):
    """Test the explanation of the ipython notebook example."""
    solution_wv = execute_arccalibration(ldebug=ldebug, lplot=lplot)

    coeff_expected = np.array([3.99875794e+03, 9.59950578e-01, 1.72739867e-05])
    assert np.allclose(solution_wv.coeff, coeff_expected)
    assert np.allclose(solution_wv.cr_linear.crval,
                       3999.7179085283897)  # 3996.42717772)
    assert np.allclose(solution_wv.cr_linear.crmin, 3999.7179085283897)
    assert np.allclose(solution_wv.cr_linear.crmax, 4999.8604201544294)
    assert np.allclose(solution_wv.cr_linear.cdelt,
                       0.97765641410170068)  # 0.978303317095)

    print("TEST: test__execute_notebook_example... OK")


# @pytest.mark.xfail
def test__execute_simple_case(ldebug=False, lplot=False):
    """Test the explanation of the ipython notebook example."""
    solution_wv = execute_arccalibration(nlines_master=15,
                                         error_xpos_arc=0.3,
                                         wv_ini_arc=3000, wv_end_arc=7000,
                                         prob_line_master_in_arc=1.0,
                                         fraction_unknown_lines=0.0,
                                         frac_triplets_for_sum=0.5,
                                         ldebug=ldebug, lplot=lplot)

    coeff_expected = np.array([2.99467778e+03, 3.89781863e+00, 1.22960881e-05])
    assert np.allclose(solution_wv.coeff, coeff_expected)
    assert np.allclose(solution_wv.cr_linear.crval,
                       2998.5756138701254)  # 2995.4384155)
    assert np.allclose(solution_wv.cr_linear.crmin, 2998.5756138701254)
    assert np.allclose(solution_wv.cr_linear.crmax, 6998.9374406492443)
    assert np.allclose(solution_wv.cr_linear.cdelt,
                       3.9104221180636549)  # 3.91231531392)

    print("TEST: test__execute_simple_case... OK")


# -----------------------------------------------------------------------------


if __name__ == '__main__':
    # test__execute_notebook_example(ldebug=True, lplot=True)
    test__execute_simple_case(ldebug=True, lplot=False)



#
# Copyright 2012-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""
Implementation of some of freedesktop.org Base Directories.

The directories are defined here:

http://standards.freedesktop.org/basedir-spec/

We only require xdg_data_dirs and xdg_config_home
"""

import os

_home = os.environ.get('HOME', '/')

xdg_data_home = os.environ.get(
    'XDG_DATA_HOME',
    os.path.join(_home, '.local', 'share')
)

xdg_data_dirs = [xdg_data_home] + \
    os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')

xdg_config_home = os.environ.get(
    'XDG_CONFIG_HOME',
    os.path.join(_home, '.config')
)

xdg_cache_home = os.environ.get(
    'XDG_CACHE_HOME',
    os.path.join(_home, '.cache')
)




import pytest

import numina.array.bbox as bbox


def test_pixin1d1():
    bb = bbox.PixelInterval1D(1, 3)
    assert bb.shape == 2
    assert bb.slice == slice(1, 3, None)


def test_pixin1d2():
    bb = bbox.PixelInterval1D.from_coordinates(1.0, 2.0)
    assert bb.shape == 2
    assert bb.slice == slice(1, 3, None)


def test_pixin1d3():
    bb = bbox.PixelInterval1D.from_coordinates(1.0, 2.0)
    bc = bbox.PixelInterval1D(1, 3)
    assert bb == bc


def test_pixin1d4():
    bb = bbox.PixelInterval1D.from_coordinates(1.0, 2.0)
    bc = bbox.PixelInterval1D(1, 4)
    assert bb != bc


def test_pixin1d5():
    bb = bbox.PixelInterval1D(2, 2)
    assert bb.shape == 0
    assert bb.slice == slice(2, 2, None)


def test_pixin1d6():
    with pytest.raises(ValueError):
        bbox.PixelInterval1D(1, 0)


def test_pixin1d7():
    with pytest.raises(TypeError):
        bbox.PixelInterval1D(1, 1.5)

    with pytest.raises(TypeError):
        bbox.PixelInterval1D(1.5, 2)


def test_pixin1d8():
    assert len(bbox.PixelInterval1D(1, 12)) == 11
    assert len(bbox.PixelInterval1D(3, 3)) == 0


def test_pixin1d9():
    assert bool(bbox.PixelInterval1D(1, 12))

    assert not bool(bbox.PixelInterval1D(3, 3))


def test_pixin1d10():
    b1 = bbox.PixelInterval1D(1, 12)
    b2 = bbox.PixelInterval1D(8, 14)
    b3 = b1.union(b2)
    assert b3 == bbox.PixelInterval1D(1, 14)

    b1 = bbox.PixelInterval1D(1, 12)
    b2 = bbox.PixelInterval1D(20, 22)
    b3 = b1.union(b2)
    assert b3 == bbox.PixelInterval1D(1, 22)


def test_pixin1d11():
    b1 = bbox.PixelInterval1D(3, 3)
    b2 = bbox.PixelInterval1D(8, 14)
    b3 = b1.union(b2)
    assert b3 == bbox.PixelInterval1D(8, 14)

    b3 = b2.union(b1)
    assert b3 == bbox.PixelInterval1D(8, 14)

    b1 = bbox.PixelInterval1D(5, 5)
    b2 = bbox.PixelInterval1D(9, 9)
    b3 = b1.union(b2)
    assert not bool(b3)


def test_pixin1d12():
    b1 = bbox.PixelInterval1D(1, 4)
    b2 = bbox.PixelInterval1D(2, 6)
    b3 = b1.intersection(b2)
    assert b3 == bbox.PixelInterval1D(2, 4)

    b3 = b2.intersection(b1)
    assert b3 == bbox.PixelInterval1D(2, 4)

    b1 = bbox.PixelInterval1D(1, 8)
    b2 = bbox.PixelInterval1D(3, 5)
    b3 = b1.intersection(b2)
    assert b3 == bbox.PixelInterval1D(3, 5)

    b3 = b2.intersection(b1)
    assert b3 == bbox.PixelInterval1D(3, 5)

    b1 = bbox.PixelInterval1D(1, 4)
    b2 = bbox.PixelInterval1D(5, 9)
    b3 = b1.intersection(b2)
    assert not bool(b3)


def test_pixin1d13():
    b1 = bbox.PixelInterval1D(1, 4)
    b2 = bbox.PixelInterval1D(2, 14)
    b3 = b1.intersection(b2)
    assert b3 == bbox.PixelInterval1D(2, 4)

    b3 = b2.intersection(b1)
    assert b3 == bbox.PixelInterval1D(2, 4)

    b1 = bbox.PixelInterval1D(5, 5)
    b2 = bbox.PixelInterval1D(9, 9)
    b3 = b1.intersection(b2)
    assert not bool(b3)


def test_pixin1d14():
    # Empty intervals are considered equal
    b1 = bbox.PixelInterval1D(1, 1)
    b2 = bbox.PixelInterval1D(2, 2)

    assert b1 == b2


def test_pixin1():
    bb = bbox.PixelInterval((2, 80), (1, 3))
    assert bb.shape == (78, 2)
    assert bb.slice == (slice(2, 80, None), slice(1, 3, None))


def test_pixin2():
    bb = bbox.PixelInterval.from_coordinates((2.0, 79.0), (1.0, 2.0))
    assert bb.shape == (78, 2)
    assert bb.slice == (slice(2, 80, None), slice(1, 3, None))


def test_pixin3():
    bb = bbox.PixelInterval.from_coordinates((2.0, 79.0), (1.0, 2.0))
    bc = bbox.PixelInterval((2, 80), (1, 3))
    assert bb == bc


def test_pixin4():
    bb = bbox.PixelInterval.from_coordinates((2.0, 79.0), (1.0, 2.0))
    bc = bbox.PixelInterval((2, 80), (1, 4))
    assert bb != bc


def test_pixin5():
    bb = bbox.PixelInterval((2, 2), (1, 1))
    assert bb.shape == (0, 0)
    assert bb.slice == (slice(2, 2, None), slice(1, 1, None))

    import numpy as np
    a = np.zeros((10, 10))
    assert (a[bb.slice].shape) == bb.shape


def test_pixin6():
    with pytest.raises(ValueError) as einfo:
        bbox.PixelInterval((1, 0), (2, 3))
    assert str(einfo.value) == "'pix2' must be >= 'pix1' in axis 0"

    with pytest.raises(ValueError) as einfo:
        bbox.PixelInterval((1, 2), (2, 1))

    assert str(einfo.value) == "'pix2' must be >= 'pix1' in axis 1"


def test_pixin7():
    with pytest.raises(TypeError) as einfo:
        bbox.PixelInterval((1, 1.5), (2, 3))

    assert str(einfo.value) == "'pix2' must be integer in axis 0"

    with pytest.raises(TypeError) as einfo:
        bbox.PixelInterval((1, 2), (2.5, 3))

    assert str(einfo.value) == "'pix1' must be integer in axis 1"


def test_pixin8():
    bb = bbox.PixelInterval((2, 80), (1, 3))
    assert len(bb) == 156

    bb = bbox.PixelInterval((2, 80), (1, 1))
    assert len(bb) == 0

    bb = bbox.PixelInterval()
    assert len(bb) == 0


def test_pixin9():
    bb = bbox.PixelInterval((2, 80), (1, 3))
    assert bool(bb)

    bb = bbox.PixelInterval((2, 80), (1, 1))
    assert not bool(bb)

    bb = bbox.PixelInterval()
    assert not bool(bb)


def test_pixin10():
    bb = bbox.PixelInterval((2, 80), (1, 3))
    assert bb.ndim == 2

    bb = bbox.PixelInterval((2, 80), (1, 1), (4, 4))
    assert bb.ndim == 3

    bb = bbox.PixelInterval()
    assert bb.ndim == 0


def test_pixin11():
    b1 = bbox.PixelInterval((1, 12), (3, 6))
    b2 = bbox.PixelInterval((8, 14), (1, 9))
    b3 = b1.union(b2)
    assert b3 == bbox.PixelInterval((1, 14), (1, 9))

    b1 = bbox.PixelInterval((1, 12), (4, 7))
    b2 = bbox.PixelInterval((20, 22), (15, 18))
    b3 = b1.union(b2)
    assert b3 == bbox.PixelInterval((1, 22), (4, 18))


def test_pixin12():
    b1 = bbox.PixelInterval((4, 9), (3, 3))
    b2 = bbox.PixelInterval((4, 9), (8, 14))
    b3 = b1.union(b2)
    assert b3 == bbox.PixelInterval((4, 9), (8, 14))

    b3 = b2.union(b1)
    assert b3 == bbox.PixelInterval((4, 9), (8, 14))

    b1 = bbox.PixelInterval((5, 5), (3, 12))
    b2 = bbox.PixelInterval((9, 9), (8, 12))
    b3 = b1.union(b2)
    assert not bool(b3)


def test_pixin13():
    b1 = bbox.PixelInterval((3, 8), (1, 4))
    b2 = bbox.PixelInterval((2, 6), (2, 6))
    b3 = b1.intersection(b2)
    assert b3 == bbox.PixelInterval((3, 6), (2, 4))

    b3 = b2.intersection(b1)
    assert b3 == bbox.PixelInterval((3, 6), (2, 4))

    b1 = bbox.PixelInterval((1, 8), (4, 9))
    b2 = bbox.PixelInterval((3, 5), (4, 9))
    b3 = b1.intersection(b2)
    assert b3 == bbox.PixelInterval((3, 5), (4, 9))

    b3 = b2.intersection(b1)
    assert b3 == bbox.PixelInterval((3, 5), (4, 9))

    b1 = bbox.PixelInterval((1, 4), (3, 6))
    b2 = bbox.PixelInterval((5, 9), (3, 7))
    b3 = b1.intersection(b2)
    assert not bool(b3)


def test_pixin14():
    b1 = bbox.PixelInterval((1, 4))
    b2 = bbox.PixelInterval((2, 14))
    b3 = b1.intersection(b2)
    assert b3 == bbox.PixelInterval((2, 4))

    b3 = b2.intersection(b1)
    assert b3 == bbox.PixelInterval((2, 4))

    b1 = bbox.PixelInterval((5, 5))
    b2 = bbox.PixelInterval((9, 9))
    b3 = b1.intersection(b2)
    assert not bool(b3)


def test_pixin15():
    """Empty intervals are considered equal"""
    b1 = bbox.PixelInterval((1, 1), (3, 4))
    b2 = bbox.PixelInterval((2, 2), (5, 6))

    assert b1 == b2


def test_pixin16():
    # Empty intervals are considered equal
    i1 = bbox.PixelInterval1D(1, 2)
    i2 = bbox.PixelInterval1D(4, 5)
    b1 = bbox.PixelInterval.from_intervals([i1, i2])
    assert b1 == bbox.PixelInterval((1, 2), (4, 5))


def test_bbox1():
    bb = bbox.BoundingBox(1, 3, 2, 80)
    assert bb.shape == (78, 2)
    assert bb.slice == (slice(2, 80, None), slice(1, 3, None))


def test_bbox2():
    bb = bbox.BoundingBox.from_coordinates(1.0, 2.0, 2.0, 79.0)
    assert bb.shape == (78, 2)
    assert bb.slice == (slice(2, 80, None), slice(1, 3, None))


def test_bbox3():
    bb = bbox.BoundingBox.from_coordinates(1.0, 2.0, 2.0, 79.0)
    bc = bbox.BoundingBox(1, 3, 2, 80)
    assert bb == bc


def test_bbox4():
    bb = bbox.BoundingBox.from_coordinates(1.0, 2.0, 2.0, 79.0)
    bc = bbox.BoundingBox(1, 3, 3, 80)
    assert bb != bc


def test_bbox5():
    bc = bbox.BoundingBox(1, 3, 3, 80)
    assert bc.extent == (0.5, 2.5, 2.5, 79.5)



#
# Copyright 2008-2019 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest

import numpy

from numina.array.imsurfit import imsurfit


def calc_ws(xx, yy):
    res = [
        1, xx, yy, xx ** 2, xx * yy, yy ** 2,
        xx ** 3, xx ** 2 * yy, xx * yy ** 2, yy ** 3,
        xx ** 4, xx ** 3 * yy, xx ** 2 * yy ** 2,
        xx * yy ** 3, yy ** 4
    ]
    return res


@pytest.mark.parametrize("steps", [100, 458])
@pytest.mark.parametrize("order, results", [
    (1, [456.0, 0.3, -0.9]),
    (2, [1.0, 0.1, 12, 3.0, -11.8, 9.2]),
    (3, [456.0, 0.3, -0.9, 0.03, -0.01, 0.07, 0.0, -10, 0.0, 0.04]),
    (4, [-11.0, -1.5, -0.1, 0.14, -15.03,
         0.07, 0.448, -0.28, 1.4, 1.24,
         -3.2, -1.2, 2.24, -8.1, -0.03])
])
def test_fit(order, results, steps):
    """Test cuartic fitting."""
    results0 = numpy.array(results)

    x0 = numpy.linspace(-1.0, 1.0, steps)
    y0 = numpy.linspace(-1.0, 1.0, steps)
    xx, yy = numpy.meshgrid(x0, y0, indexing='ij')
    # xx, yy = numpy.mgrid[-1:1:100j, -1:1:100j]
    ws0 = calc_ws(xx, yy)

    dd = [cf * v for cf, v in zip(results0.flat, ws0)]
    z = sum(dd)
    results, = imsurfit(z, order=order)
    assert numpy.allclose(results0, results)




import warnings

import pytest

from ..arccalibration import WavecalFeature


features = [
    {'category': 'A', 'xpos': 100.0, 'reference': 3210.0,
     'wavelength': 3100.0, 'line_ok': True, 'ypos': 0.0, 'peak': 0.0,
     'funcost': 12.0, 'fwhm': 0.0,
     'lineid': 1},
    {'category': 'A', 'xpos': 150.0, 'reference': 3310.0,
     'wavelength': 3150.0, 'line_ok': True, 'ypos': 0.0, 'peak': 0.0,
     'funcost': 12.0, 'fwhm': 0.0,
     'lineid': 2},
    {'category': 'C', 'xpos': 250.0, 'reference': 3410.0,
     'wavelength': 3250.0, 'line_ok': True, 'ypos': 0.0, 'peak': 0.0,
     'funcost': 13.0, 'fwhm': 0.0,
     'lineid': 3},
    {'category': 'X', 'xpos': 250.0, 'reference': 3410.0,
     'wavelength': 3250.0, 'line_ok': True, 'ypos': 0.0, 'peak': 0.0,
     'funcost': float('inf'), 'fwhm': 0.0,
     'lineid': 4}
]


def create_features(orig):

    # Create Features
    features = []
    for feature in orig['features']:
        m = WavecalFeature(**feature)
        features.append(m)
    return features


def test_serialize1():

    feature0 = features[0]

    mm = WavecalFeature(**feature0)
    assert feature0 == mm.__getstate__()


def test_serialize2():
    warnings.simplefilter('always')

    feature3 = features[3]
    mm = WavecalFeature(**feature3)
    feature3_c = feature3.copy()
    feature3_c['funcost'] = 1e50
    assert feature3_c == mm.__getstate__()

    with pytest.warns(RuntimeWarning):
        mm.__getstate__()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Apply ADR correction using offsets from FITS extension"""
import argparse

import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from datetime import datetime
import numpy.ma as ma
from reproject.mosaicking import find_optimal_celestial_wcs
from reproject import reproject_interp, reproject_adaptive, reproject_exact
import sys

from .ctext import ctext

REPROJECT_METHODS = ['interp', 'adaptive', 'exact']


def apply_adr_correction_from_extension_in_3d_cube(
        hdul,
        extname_adr,
        extname_mask,
        final_celestial_wcs,
        reproject_method,
        verbose=False,
):
    """Apply ADR correction using offsets from FITS extension.

    Parameters
    ----------
    input_hdul : `astropy.io.fits.HDUList`
        Input HDUList instance.
    extname_adr : str
        FITS extension name with ADR correction data.
    extname_mask : str
        FITS extension name with mask. If 'None', use np.nan in
        image data.
    final_celestial_wcs : str, file-like, `pathlib.Path` or None
        FITS filename with final celestial WCS. If None,
        compute output celestial WCS for current 3D cube.
    reproject_method : str
        Reprojection method. See 'REPROJECT_METHODS' above.
    verbose : bool
        If True, display additional information.

    Returns
    -------
    output_hdul : `astropy.io.fits.HDUList`
        Instance of HDUList with two HDU:
        - PRIMARY: corrected image
        - MASK: array with masked pixels
    """
    primary_header = hdul[0].header
    if primary_header["NAXIS"] != 3:
        raise ValueError("Expected NAXIS=3 not found in PRIMARY HDU")
    # read 3D data cube
    data3d = hdul[0].data.copy()
    # read ADR data
    if extname_adr not in hdul:
        raise ValueError(f"Extension '{extname_adr}' not found in FITS file.")
    else:
        if verbose:
            print(f'Reading ADR correction from {extname_adr} extension')
        table_adrcross = hdul[extname_adr].data.copy()
    # read mask or generate one from np.nan
    if extname_mask is None:
        hdu3d_mask = None
    else:
        if extname_mask not in hdul:
            raise ValueError(f"Extension '{extname_mask}' not found in FITS file.")
        else:
            if verbose:
                print(f'Reading mask from {extname_mask} extension')
            hdu3d_mask = hdul[extname_mask]
            bitpix_mask = hdu3d_mask.header['BITPIX']
            if bitpix_mask != 8:
                raise ValueError(f"BITPIX (mask): {bitpix_mask} is not 8")
            if data3d.shape != hdu3d_mask.data.shape:
                raise ValueError(f"Shape of PRIMARY and {extname_mask} are different")
    # generate mask from np.nan when necessary
    if hdu3d_mask is None:
        if verbose:
            print("Generating mask from np.nan in PRIMARY HDU")
        mask3d = np.isnan(data3d).astype(np.uint8)
    else:
        mask3d = hdu3d_mask.data.copy()
    num_masked_pixels_in_data3d = mask3d.sum()
    if verbose:
        print(f"Number of masked pixels in the input 3D array: {num_masked_pixels_in_data3d}")

    # input 3D WCS
    wcs3d = WCS(primary_header)
    naxis1, naxis2, naxis3 = wcs3d.pixel_shape

    # ADR at the center of the field of view
    delta_x_center_ifu = table_adrcross['Delta_x']
    delta_y_center_ifu = table_adrcross['Delta_y']
    x_center_ifu, y_center_ifu = wcs3d.celestial.wcs.crpix  # FITS convention
    center_ifu_coord = wcs3d.celestial.pixel_to_world(
        x_center_ifu - 1.0,  # Python convention
        y_center_ifu - 1.0   # Python convention
    )
    if verbose:
        print(f'Center IFU coord: {center_ifu_coord}')
    x_center_ifu_corrected = x_center_ifu + delta_x_center_ifu
    y_center_ifu_corrected = y_center_ifu + delta_y_center_ifu

    # optimal celestial WCS for corrected cube
    if final_celestial_wcs is None:
        # compute final celestial WCS for this 3D cube
        wcs2d_blue = wcs3d.celestial.deepcopy()
        wcs2d_blue.wcs.crpix = np.array([x_center_ifu_corrected[0], y_center_ifu_corrected[0]])
        if verbose:
            print(f"\nwcs2d_blue:\n{wcs2d_blue}")
        wcs2d_red = wcs3d.celestial.deepcopy()
        wcs2d_red.wcs.crpix = np.array([x_center_ifu_corrected[-1], y_center_ifu_corrected[-1]])
        if verbose:
            print(f"\nwcs2d_red:\n{wcs2d_red}")
        wcs_mosaic2d, shape_mosaic2d = find_optimal_celestial_wcs(
            input_data=[
                ((naxis2, naxis1), wcs2d_blue),
                ((naxis2, naxis1), wcs2d_red)
            ],
            auto_rotate=False
        )
    else:
        # make use of an external celestial WCS projection
        with fits.open(final_celestial_wcs) as hdul_mosaic2d:
            wcs_mosaic2d = WCS(hdul_mosaic2d[0].header)
            shape_mosaic2d = hdul_mosaic2d[0].header['NAXIS2'], hdul_mosaic2d[0].header['NAXIS1']
    if verbose:
        print(f"\nwcs_mosaic2d:\n{wcs_mosaic2d}")
        print(f"shape_mosaic2d: {shape_mosaic2d}")

    # initialize corrected 3D cube (using a masked array!)
    naxis3_corrected3d = naxis3
    naxis2_corrected3d, naxis1_corrected3d = shape_mosaic2d
    if verbose:
        print("NAXIS1, NAXIS2, NAXIS3 of corrected 3D cube: "
              f"{naxis1_corrected3d}, {naxis2_corrected3d}, {naxis3_corrected3d}")
    shape_corrected3d = (naxis3_corrected3d, naxis2_corrected3d, naxis1_corrected3d)
    array3d_corrected = ma.array(np.zeros(shape_corrected3d),
                                 mask=np.full(shape_corrected3d, fill_value=False))

    # compute corrected 3D cube slice by slice
    wcs2d = wcs3d.celestial.deepcopy()
    if verbose:
        print(f"Reprojection method: {reproject_method} (please wait...)")
    time_ini = datetime.now()
    """
    # alternative code for debugging purposes
    range1 = range(50)
    range2 = range(naxis3-50, naxis3)
    combined_range = list(x for r in (range1, range2) for x in r)
    for k in combined_range:
    """
    for k in range(naxis3):
        wcs2d.wcs.crpix = np.array([x_center_ifu_corrected[k], y_center_ifu_corrected[k]])
        data2d = data3d[k, :, :]
        hdu2d_data = fits.PrimaryHDU(data2d)
        hdu2d_data.header.extend(wcs2d.to_header(), update=True)
        mask2d = mask3d[k, :, :]
        hdu2d_mask = fits.PrimaryHDU(mask2d)
        hdu2d_mask.header.extend(wcs2d.to_header(), update=True)
        if reproject_method == "interp":
            data2d_resampled, footprint_data2d_resampled = reproject_interp(
                input_data=hdu2d_data,
                output_projection=wcs_mosaic2d,
                shape_out=shape_mosaic2d,
            )
            if num_masked_pixels_in_data3d > 0:
                mask2d_resampled, _ = reproject_interp(
                    input_data=hdu2d_mask,
                    output_projection=wcs_mosaic2d,
                    shape_out=shape_mosaic2d,
                )
            else:
                mask2d_resampled = None
        elif reproject_method == "adaptive":
            data2d_resampled, footprint_data2d_resampled = reproject_adaptive(
                input_data=hdu2d_data,
                output_projection=wcs_mosaic2d,
                shape_out=shape_mosaic2d,
                conserve_flux=True,
                kernel='Gaussian'
            )
            if num_masked_pixels_in_data3d > 0:
                mask2d_resampled, _ = reproject_adaptive(
                    input_data=hdu2d_mask,
                    output_projection=wcs_mosaic2d,
                    shape_out=shape_mosaic2d,
                    conserve_flux=True,
                    kernel='Gaussian'
                )
            else:
                mask2d_resampled = None
        elif reproject_method == "exact":
            data2d_resampled, footprint_data2d_resampled = reproject_exact(
                input_data=hdu2d_data,
                output_projection=wcs_mosaic2d,
                shape_out=shape_mosaic2d,
            )
            if num_masked_pixels_in_data3d > 0:
                mask2d_resampled, _ = reproject_exact(
                    input_data=hdu2d_mask,
                    output_projection=wcs_mosaic2d,
                    shape_out=shape_mosaic2d,
                )
            else:
                mask2d_resampled = None
        else:
            raise ValueError(f"Reprojection method '{reproject_method}' not recognized.")
        array3d_corrected[k, :, :] = data2d_resampled
        if mask2d_resampled is None:
            # use only the footprint
            array3d_corrected[k, :, :].mask = (footprint_data2d_resampled == 0)
        else:
            # merge corrected mask with footprint
            array3d_corrected[k, :, :].mask = np.logical_or(
                np.logical_or(mask2d_resampled, mask2d_resampled==np.nan),
                (footprint_data2d_resampled == 0)
            )

    time_end = datetime.now()

    # 3D footprint
    footprint3d_corrected = np.ones_like(array3d_corrected)
    footprint3d_corrected[np.isnan(array3d_corrected)] = 0

    if verbose:
        print(f"Reprojection finished! (elapsed time: {(time_end - time_ini).total_seconds()} seconds)")
        print("\nFlux check:")
        flux1 = np.sum(data3d)
        flux2 = np.nansum(array3d_corrected)
        print(f"- total counts in original  3D cube: {flux1}")
        print(f"- total counts in corrected 3D cube: {flux2}")
        print(f"- ratio original/corrected.........: {flux1/flux2}")
        print(f"\nFootprint coverage (fraction): {np.sum(footprint3d_corrected) / footprint3d_corrected.size}")

    # generate single 3D WCS combining the celestial and spectral axes
    header3d_corrected = wcs_mosaic2d.to_header()
    header_spectral = wcs3d.spectral.to_header()
    header3d_corrected['WCSAXES'] = 3
    for item in ['CRPIX', 'CDELT', 'CUNIT', 'CTYPE', 'CRVAL']:
        # insert {item}3 after {item}2 to preserve the order in the header
        header3d_corrected.insert(
            f'{item}2',
            (f'{item}3', header_spectral[f'{item}1'], header_spectral.comments[f'{item}1']),
            after=True)
    # IMPORTANT: CDELT3 needs to be recomputed
    # (one option is to use PC3_3 if CDELT3=1; another one is to compute 1 pixel increment)
    header3d_corrected['CDELT3'] =  wcs3d.spectral.pixel_to_world(2).value - wcs3d.spectral.pixel_to_world(1).value
    if verbose:
        print("\nheader3d_corrected:")
        for line in header3d_corrected.cards:
            print(line)

    # generate result
    hdu = fits.PrimaryHDU(array3d_corrected.data.astype(np.float32))
    hdu.header.update(header3d_corrected)
    hdu_mask = fits.ImageHDU(data=array3d_corrected.mask.astype(np.uint8))
    hdu_mask.header['EXTNAME'] = 'MASK'
    hdu_mask.header.update(header3d_corrected)
    output_hdul = fits.HDUList([hdu, hdu_mask])

    return output_hdul


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(description="Compare ADR extensions in 3D cube")
    parser.add_argument("inputfile", help="Input 3D FITS file", type=str)
    parser.add_argument("--extname_adr", help="Name of the extension with ADR correction",
                        type=str, default=None)
    parser.add_argument("--extname_mask",
                        help="Name of the extension with mask in input 3D cube. "
                        "Default 'None': use np.nan in image",
                        default=None, type=str)
    parser.add_argument("--final_celestial_wcs",
                        help="Final celestial WCS projection. Default None (compute for current 3D cube)",
                        type=str, default=None)
    parser.add_argument("--output", help="Output filename",
                        type=str, default=None)
    parser.add_argument('--reproject_method',
                        help='Reprojection method (interp, adaptive, exact)',
                        type=str, choices=REPROJECT_METHODS, default='adaptive')
    parser.add_argument("--verbose",
                        help="Display intermediate information",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.extname_adr is None:
        raise ValueError("You must specify an extension name with --extname_adr")
    else:
        extname_adr = args.extname_adr.upper()

    extname_mask = args.extname_mask
    if extname_mask is not None:
        if extname_mask.lower() in ['none', 'nan']:
            extname_mask = None
        else:
            extname_mask = extname_mask.upper()

    if args.output is None:
        raise ValueError("You must specify an output filename with --output")

    with fits.open(args.inputfile) as input_hdul:
        if args.verbose:
            print(input_hdul.info())
        output_hdul = apply_adr_correction_from_extension_in_3d_cube(
            hdul=input_hdul,
            extname_adr=extname_adr,
            extname_mask=extname_mask,
            final_celestial_wcs=args.final_celestial_wcs,
            reproject_method=args.reproject_method,
            verbose=args.verbose
        )

    # save result
    if args.verbose:
        print(f"\nSaving file: {args.output}")
    output_hdul.writeto(args.output, overwrite=True)


if __name__ == '__main__':
    main()
